/*
 * Transcoder.Native Transcode.h
 * Copyright (C) 2007-2008 doubleTwist Corporation
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once
 
#ifdef _WIN32
	#ifdef FFMPEG_EXPORT_SYMBOLS
		#define DLL_EXPORT __declspec(dllexport)
	#elif defined FFMPEG_IMPORT_SYMBOLS
		#define DLL_EXPORT __declspec(dllimport)
	#else
		#define DLL_EXPORT
	#endif
#else
	#define DLL_EXPORT
	#define __int64 int64_t
	#define _strtoi64 strtoll
	#define _snprintf snprintf
	#define HAVE_LRINT 1
	#define HAVE_LRINTF 1
#endif

#include <vector>
#include <string>

namespace transcode
{

enum MediaType { eMediaType_Unknown, eMediaType_Video, eMediaType_Audio };

enum FileFormat {	eFileFormat_Unknown, eFileFormat_MP4, eFileFormat_ASF, eFileFormat_AVI, eFileFormat_FLV,
					eFileFormat_MPEG, eFileFormat_WAV };

enum VideoEncoding {	eVideoEncoding_Unknown, eVideoEncoding_MP4, eVideoEncoding_H263, eVideoEncoding_H264,
						eVideoEncoding_WMV7, eVideoEncoding_WMV8, eVideoEncoding_WMV9, eVideoEncoding_VC1,
						eVideoEncoding_DVVideo, eVideoEncoding_MPEG1, eVideoEncoding_MPEG2,
						eVideoEncoding_H263Spark, eVideoEncoding_VP6, eVideoEncoding_XVID };

enum AudioEncoding {	eAudioEncoding_Unknown, eAudioEncoding_AMR_NB, eAudioEncoding_AMR_WB, eAudioEncoding_AAC,
						eAudioEncoding_AC3, eAudioEncoding_MP3, eAudioEncoding_WMA1, eAudioEncoding_WMA2, eAudioEncoding_DVAudio,
						eAudioEncoding_PCM };

class Progress
{
public:
	DLL_EXPORT Progress();
	DLL_EXPORT virtual ~Progress();

	DLL_EXPORT virtual bool Update(float percentComplete) = 0;
};

class VideoMetadata
{
public:
	DLL_EXPORT VideoMetadata();
	DLL_EXPORT ~VideoMetadata();

	DLL_EXPORT VideoEncoding Encoding() const;
	void Encoding(VideoEncoding);

	DLL_EXPORT float Duration() const;
	void Duration(float);

	DLL_EXPORT long AverageBitrate() const;
	void AverageBitrate(long);

	DLL_EXPORT int Width() const;
	void Width(int);

	DLL_EXPORT int Height() const;
	void Height(int);

	DLL_EXPORT float FrameRate() const;
	void FrameRate(float);

	DLL_EXPORT float PixelAspectRatio() const;
	void PixelAspectRatio(float);

	DLL_EXPORT int Profile() const;
	void Profile(int);

	DLL_EXPORT int Level() const;
	void Level(int);

protected:
	int mWidth;
	int mHeight;
	long mAverageBitrate;
	float mFrameRate;
	float mDuration;
	float mPixelAspectRatio;
	VideoEncoding mEncoding;
	int mProfile;
	int mLevel;
};

class AudioMetadata
{
public:
	DLL_EXPORT AudioMetadata();
	DLL_EXPORT ~AudioMetadata();

	DLL_EXPORT AudioEncoding Encoding() const;
	void Encoding(AudioEncoding);

	DLL_EXPORT float Duration() const;
	void Duration(float);

	DLL_EXPORT int Bitrate() const;
	void Bitrate(int);

	DLL_EXPORT int SampleRate() const;
	void SampleRate(int);

	DLL_EXPORT int SampleSize() const;
	void SampleSize(int);

	DLL_EXPORT int ChannelCount() const;
	void ChannelCount(int);

protected:
	AudioEncoding mEncoding;
	float mDuration;
	int mBitrate;
	int mSampleSize;
	int mSampleRate;
	int mChannelCount;
};

class Metadata
{
public:
	DLL_EXPORT Metadata();
	DLL_EXPORT ~Metadata();

	//DLL_EXPORT Metadata(const Metadata& rhs);

	DLL_EXPORT MediaType Type() const;
	void Type(MediaType);

	DLL_EXPORT FileFormat Format() const;
	void Format(FileFormat);

	DLL_EXPORT float Duration() const;
	void Duration(float);

	DLL_EXPORT long AverageBitrate() const;
	void AverageBitrate(long);

	DLL_EXPORT bool HasAudio() const;
	void HasAudio(bool);

	DLL_EXPORT std::string Title() const;
	void Title(const std::string&);

	DLL_EXPORT std::string Artist() const;
	void Artist(const std::string&);

	DLL_EXPORT std::string Album() const;
	void Album(const std::string&);

	DLL_EXPORT std::string Genre() const;
	void Genre(const std::string&);

	DLL_EXPORT VideoMetadata& VideoMetaData();
	DLL_EXPORT AudioMetadata& AudioMetaData();

protected:
	MediaType mMediaType;
	FileFormat mFileFormat;

	float mDuration;
	long mAverageBitrate;
	bool mHasAudio;

	std::string mTitle;
	std::string mArtist;
	std::string mAlbum;
	std::string mGenre;

	transcode::VideoMetadata mVideoMetadata;
	transcode::AudioMetadata mAudioMetadata;
};
 
DLL_EXPORT void Transcode(const std::vector<std::string>& argv, Progress& progress);
DLL_EXPORT Metadata GetMetadata(const std::string& filePath);
DLL_EXPORT void SetLogFile(const std::string& filePath);

}
