/*
 * Part of this file is from FFmpeg and under the LGPL
 * Copyright (c) 2000-2003 Fabrice Bellard
 *
 * FFmpeg is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * FFmpeg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FFmpeg; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/*
 * Part of this file is written by doubleTwist and under the GPL
 * Copyright (C) 2007-2008 doubleTwist Corporation
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "Transcode.h"

/* needed for usleep() */
#define _XOPEN_SOURCE 500

#define INT64_MAX INT64_C(9223372036854775807)

extern "C"
{
#include <errno.h>
#include <signal.h>
#include <unistd.h>
#include "libavformat/avformat.h"
#include "libavdevice/avdevice.h"
#include "libswscale/swscale.h"
#include "libavcodec/opt.h"
#include "libavcodec/audioconvert.h"
#include "libavutil/fifo.h"
#include "libavutil/avstring.h"
#include "float_cast.h"
}

#ifdef _WIN32
#include "Windows.h"
#include "Winbase.h"
#endif

#include <exception>

#include "LogFileOpener.h"

#ifdef HAVE_SYS_RESOURCE_H
#include <sys/types.h>
#include <sys/resource.h>
#elif defined(HAVE_GETPROCESSTIMES)
#include <windows.h>
#endif

#if defined(HAVE_TERMIOS_H)
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <termios.h>
#elif defined(HAVE_CONIO_H)
#include <conio.h>
#endif
#undef time //needed because HAVE_AV_CONFIG_H is defined on top
#include <time.h>

#include <sstream>

#include "cmdutils.h"

#undef NDEBUG
#include <assert.h>

#if !defined(INFINITY) && defined(HUGE_VAL)
#define INFINITY HUGE_VAL
#endif

#undef exit

class Options;

/* select an input stream for an output stream */
typedef struct AVStreamMap {
    int file_index;
    int stream_index;
    int sync_file_index;
    int sync_stream_index;
} AVStreamMap;

/** select an input file for an output file */
typedef struct AVMetaDataMap {
    int out_file;
    int in_file;
} AVMetaDataMap;

extern const OptionDef options[];

static int opt_default(const char *opt, const char *arg, Options& options);

#define MAX_FILES 20
#define QSCALE_NONE -99999
#define DEFAULT_PASS_LOGFILENAME "ffmpeg2pass"
#define INT64_C(x) x##LL
#define snprintf _snprintf

struct AVInputStream;
struct DtStream;

typedef struct AVOutputStream {
    int file_index;          /* file index */
    int index;               /* stream index in the output file */
    int source_index;        /* AVInputStream index */
    AVStream *st;            /* stream in the output file */
    int encoding_needed;     /* true if encoding needed for this stream */
    int frame_number;
    /* input pts and corresponding output pts
       for A/V sync */
    //double sync_ipts;        /* dts from the AVPacket of the demuxer in second units */
    struct AVInputStream *sync_ist; /* input stream to sync against */
    int64_t sync_opts;       /* output frame counter, could be changed to some true timestamp */ //FIXME look at frame_number
    /* video only */
    int video_resample;
    AVFrame pict_tmp;      /* temporary image for resampling */
    struct SwsContext *img_resample_ctx; /* for image resampling */
    int resample_height;

    int video_crop;
    int topBand;             /* cropping area sizes */
    int leftBand;

    int video_pad;
    int padtop;              /* padding area sizes */
    int padbottom;
    int padleft;
    int padright;

    /* audio only */
    int audio_resample;
    ReSampleContext *resample; /* for audio resampling */
    int reformat_pair;
    AVAudioConvert *reformat_ctx;
    AVFifoBuffer fifo;     /* for compression: one audio fifo per codec */
    FILE *logfile;
} AVOutputStream;

typedef struct AVInputStream {
    int file_index;
    int index;
    AVStream *st;
    int discard;             /* true if stream data should be discarded */
    int decoding_needed;     /* true if the packets must be decoded in 'raw_fifo' */
    int64_t sample_index;      /* current sample */

    int64_t       start;     /* time when read started */
    unsigned long frame;     /* current frame */
    int64_t       next_pts;  /* synthetic pts for cases where pkt.pts
                                is not defined */
    int64_t       pts;       /* current pts */
    int is_start;            /* is 1 at the start and after a discontinuity */
} AVInputStream;

typedef struct AVInputFile {
    int eof_reached;      /* true if eof reached */
    int ist_index;        /* index of first stream in ist_table */
    int buffer_size;      /* current total buffer size */
    int nb_streams;       /* nb streams we are aware of */
} AVInputFile;

#ifdef HAVE_TERMIOS_H

/* init terminal so that we can grab keys */
static struct termios oldtty;
#endif

using transcode::Progress;

Progress::Progress()
{
}

Progress::~Progress()
{
}


class Options
{
public:
	AVFormatContext *input_files[MAX_FILES];
	int64_t input_files_ts_offset[MAX_FILES];
    double input_files_ts_scale[MAX_FILES][MAX_STREAMS];
	int nb_input_files;

	AVFormatContext *output_files[MAX_FILES];
	int nb_output_files;

	AVStreamMap stream_maps[MAX_FILES*MAX_STREAMS];
	int nb_stream_maps;

	AVMetaDataMap meta_data_maps[MAX_FILES];
	int nb_meta_data_maps;

	AVInputFormat *file_iformat;
	AVOutputFormat *file_oformat;

	int frame_width;
	int frame_height;
	float frame_aspect_ratio;
	enum PixelFormat frame_pix_fmt;
    enum SampleFormat audio_sample_fmt;
	int frame_padtop;
	int frame_padbottom;
	int frame_padleft;
	int frame_padright;
	int padcolor[3];
	int frame_topBand;
	int frame_bottomBand;
	int frame_leftBand;
	int frame_rightBand;
	int max_frames[4];
	AVRational frame_rate;
	float video_qscale;
	uint16_t *intra_matrix;
	uint16_t *inter_matrix;

	char *video_rc_override_string;
	int video_disable;
	int video_discard;
	char *video_codec_name;
	int video_codec_tag;
	int same_quality;
	int do_deinterlace;
	int top_field_first;
	int me_threshold;
	int intra_dc_precision;
	int loop_input;
	int loop_output;
	int qp_hist;

	int intra_only;
	int audio_sample_rate;
	float audio_qscale;
	int audio_disable;
	int audio_channels;
	char  *audio_codec_name;
	int audio_codec_tag;
	char *audio_language;

	int subtitle_disable;
	char *subtitle_codec_name;
	char *subtitle_language;

	float mux_preload;
	float mux_max_delay;

	int64_t recording_time;
	int64_t start_time;
	int64_t rec_timestamp;
	int64_t input_ts_offset;
	int file_overwrite;
	char *str_title;
	char *str_author;
	char *str_copyright;
	char *str_comment;
    char *str_genre;
	char *str_album;
	int do_benchmark;
	int do_hex_dump;
	int do_pkt_dump;
	int do_psnr;
	int do_pass;
	char *pass_logfilename;
	int audio_stream_copy;
	int video_stream_copy;
	int subtitle_stream_copy;
	int video_sync_method;
	int audio_sync_method;
	float audio_drift_threshold;
	int copy_ts;
	int opt_shortest;
	int video_global_header;
	char *vstats_filename;
	FILE *fvstats;
    int opt_programid;

	int rate_emu;

	int  video_channel;
	char *video_standard;

	int audio_volume;

	int using_stdin;
	int using_vhook;
	int verbose;
	int thread_count;
	int q_pressed;
	int64_t video_size;
	int64_t audio_size;
	int64_t extra_size;
	int nb_frames_dup;
	int nb_frames_drop;
	int input_sync;
	uint64_t limit_filesize; //

	int pgmyuv_compatibility_hack;
	float dts_delta_threshold;

	int sws_flags;

	const char **opt_names;
	int opt_name_count;
	AVCodecContext *avctx_opts[CODEC_TYPE_NB];
	AVFormatContext *avformat_opts;
	struct SwsContext *sws_opts;
	int64_t timer_start;

	AVBitStreamFilterContext *video_bitstream_filters;
	AVBitStreamFilterContext *audio_bitstream_filters;
	AVBitStreamFilterContext *bitstream_filters[MAX_FILES][MAX_STREAMS];

public:
	Options()
	{
		nb_input_files = 0;
		nb_output_files = 0;
		nb_stream_maps = 0;
		nb_meta_data_maps = 0;
		file_iformat = 0;
		file_oformat = 0;
		frame_width  = 0;
		frame_height = 0;
		frame_aspect_ratio = 0;
		frame_pix_fmt = PIX_FMT_NONE;
        audio_sample_fmt = SAMPLE_FMT_NONE;
        
		frame_padtop  = 0;
		frame_padbottom = 0;
		frame_padleft  = 0;
		frame_padright = 0;

		padcolor[0] = 16; /* default to black */
		padcolor[1] = 128;
		padcolor[2] = 128;

		frame_topBand  = 0;
		frame_bottomBand = 0;
		frame_leftBand  = 0;
		frame_rightBand = 0;
		max_frames[0] = INT_MAX;
		max_frames[1] = INT_MAX;
		max_frames[2] = INT_MAX;
		max_frames[3] = INT_MAX;

		frame_rate.num = 25;
		frame_rate.den = 1;

		video_qscale = 0;
		intra_matrix = NULL;
		inter_matrix = NULL;

		video_rc_override_string=NULL;
		video_disable = 0;
		video_discard = 0;
		video_codec_name = NULL;
		video_codec_tag = 0;
		same_quality = 0;
		do_deinterlace = 0;
		top_field_first = -1;
		me_threshold = 0;
		intra_dc_precision = 8;
		loop_input = 0;
		loop_output = AVFMT_NOOUTPUTLOOP;
		qp_hist = 0;

		intra_only = 0;
		audio_sample_rate = 44100;
		audio_qscale = QSCALE_NONE;
		audio_disable = 0;
		audio_channels = 1;
		audio_codec_name = NULL;
		audio_codec_tag = 0;
		audio_language = NULL;

		subtitle_disable = 0;
		subtitle_codec_name = NULL;
		subtitle_language = NULL;

		mux_preload= 0.5;
		mux_max_delay= 0.7f;

		recording_time = INT64_MAX;
		start_time = 0;
		rec_timestamp = 0;
		input_ts_offset = 0;
		file_overwrite = 0;
		str_title = NULL;
		str_author = NULL;
		str_copyright = NULL;
		str_comment = NULL;
        str_genre = NULL;
		str_album = NULL;
		do_benchmark = 0;
		do_hex_dump = 0;
		do_pkt_dump = 0;
		do_psnr = 0;
		do_pass = 0;
		pass_logfilename = NULL;
		audio_stream_copy = 0;
		video_stream_copy = 0;
		subtitle_stream_copy = 0;
		video_sync_method= -1;
		audio_sync_method= 0;
		audio_drift_threshold= 0.1f;
		copy_ts= 0;
		opt_shortest = 0; //
		video_global_header = 0;
		vstats_filename = 0;
		fvstats = 0;
        opt_programid = 0;

		rate_emu = 0;

		video_channel = 0;
		video_standard = 0;

		audio_volume = 256;

		using_stdin = 0;
		using_vhook = 0;
		verbose = 1;
		thread_count= 1;
		q_pressed = 0;
		video_size = 0;
		audio_size = 0;
		extra_size = 0;
		nb_frames_dup = 0;
		nb_frames_drop = 0;
		input_sync = 0;
		limit_filesize = 0;

		pgmyuv_compatibility_hack=0;
		dts_delta_threshold = 10;

		sws_flags = SWS_BICUBIC;

		opt_names = 0;
		opt_name_count = 0;
		avformat_opts = 0;
		sws_opts = 0;
		timer_start = 0;

		video_bitstream_filters=0;
		audio_bitstream_filters=0;
	}

	virtual ~Options()
	{
		// close files
		for(int i=0;i<nb_output_files;i++) {
			// maybe av_close_output_file ???
			AVFormatContext *s = output_files[i];
			if (!(s->oformat->flags & AVFMT_NOFILE))
				url_fclose(s->pb);
			for(unsigned int j=0;j<s->nb_streams;j++) {
                av_metadata_free(&s->streams[j]->metadata);
				av_free(s->streams[j]->codec);
				av_free(s->streams[j]);
			}
            for(unsigned int j=0;j<s->nb_programs;j++) {
                av_metadata_free(&s->programs[j]->metadata);
            }
            for(unsigned int j=0;j<s->nb_chapters;j++) {
                av_metadata_free(&s->chapters[j]->metadata);
            }
            av_metadata_free(&s->metadata);
			av_free(s);
		}
		for(int i=0;i<nb_input_files;i++)
			av_close_input_file(input_files[i]);

		av_free(intra_matrix);
		av_free(inter_matrix);

		if (fvstats)
			fclose(fvstats);
		av_free(vstats_filename);

		av_free(opt_names);

		av_free(video_codec_name);
		av_free(audio_codec_name);
		av_free(subtitle_codec_name);

		av_free(video_standard);
    }
};

static volatile sig_atomic_t received_sigterm = 0;

static void ThrowInternalError()
{
	std::stringstream message;
	message << "File: " << __FILE__ << ", Line: " << __LINE__;
    throw std::exception(); // std::exception(message.str().c_str());
}

static void
sigterm_handler(int sig)
{
    received_sigterm = sig;
    //term_exit();
}

static void term_init(void)
{
#ifdef HAVE_TERMIOS_H
    struct termios tty;

    tcgetattr (0, &tty);
    oldtty = tty;

    tty.c_iflag &= ~(IGNBRK|BRKINT|PARMRK|ISTRIP
                          |INLCR|IGNCR|ICRNL|IXON);
    tty.c_oflag |= OPOST;
    tty.c_lflag &= ~(ECHO|ECHONL|ICANON|IEXTEN);
    tty.c_cflag &= ~(CSIZE|PARENB);
    tty.c_cflag |= CS8;
    tty.c_cc[VMIN] = 1;
    tty.c_cc[VTIME] = 0;

    tcsetattr (0, TCSANOW, &tty);
    signal(SIGQUIT, sigterm_handler); /* Quit (POSIX).  */
#endif

    signal(SIGINT , sigterm_handler); /* Interrupt (ANSI).  */
    signal(SIGTERM, sigterm_handler); /* Termination (ANSI).  */
    /*
    register a function to be called at normal program termination
    */
    //atexit(term_exit);
#ifdef CONFIG_BEOS_NETSERVER
    fcntl(0, F_SETFL, fcntl(0, F_GETFL) | O_NONBLOCK);
#endif
}

/* read a key without blocking */
static int read_key(void)
{
#if defined(HAVE_TERMIOS_H)
    int n = 1;
    unsigned char ch;
#ifndef CONFIG_BEOS_NETSERVER
    struct timeval tv;
    fd_set rfds;

    FD_ZERO(&rfds);
    FD_SET(0, &rfds);
    tv.tv_sec = 0;
    tv.tv_usec = 0;
    n = select(1, &rfds, NULL, NULL, &tv);
#endif
    if (n > 0) {
        n = read(0, &ch, 1);
        if (n == 1)
            return ch;

        return n;
    }
#elif defined(HAVE_CONIO_H)
    if(kbhit())
        return(getch());
#endif
    return -1;
}

static int decode_interrupt_cb(void)
{
	return 0;
    //return q_pressed || (q_pressed = read_key() == 'q');
}

static int read_ffserver_streams(AVFormatContext *s, const char *filename, Options& options)
{
    int err;
    AVFormatContext *ic;
    int nopts = 0;

    err = av_open_input_file(&ic, filename, NULL, FFM_PACKET_SIZE, NULL);
    if (err < 0)
        return err;
    /* copy stream format */
    s->nb_streams = ic->nb_streams;
    for(unsigned int i=0;i<ic->nb_streams;i++) {
        AVStream *st;

        // FIXME: a more elegant solution is needed
        st = (AVStream*)av_mallocz(sizeof(AVStream));
        memcpy(st, ic->streams[i], sizeof(AVStream));
        st->codec = avcodec_alloc_context();
        memcpy(st->codec, ic->streams[i]->codec, sizeof(AVCodecContext));
        s->streams[i] = st;
        

        if (st->codec->codec_type == CODEC_TYPE_AUDIO && options.audio_stream_copy)
            st->stream_copy = 1;
        else if (st->codec->codec_type == CODEC_TYPE_VIDEO && options.video_stream_copy)
            st->stream_copy = 1;

        if(!st->codec->thread_count)
            st->codec->thread_count = 1;
        if(st->codec->thread_count>1)
            avcodec_thread_init(st->codec, st->codec->thread_count);

        if(st->codec->flags & CODEC_FLAG_BITEXACT)
            nopts = 1;
    }

    if (!nopts)
        s->timestamp = av_gettime();

    av_close_input_file(ic);
    return 0;
}

static double
get_sync_ipts(const AVOutputStream *ost, const Options& options)
{
    const AVInputStream *ist = ost->sync_ist;
    return (double)(ist->pts - options.start_time)/AV_TIME_BASE;
}

static void write_frame(AVFormatContext *s, AVPacket *pkt, AVCodecContext *avctx, AVBitStreamFilterContext *bsfc){
    while(bsfc){
        AVPacket new_pkt= *pkt;
        int a= av_bitstream_filter_filter(bsfc, avctx, NULL,
                                          &new_pkt.data, &new_pkt.size,
                                          pkt->data, pkt->size,
                                          pkt->flags & PKT_FLAG_KEY);
        if(a){
            av_free_packet(pkt);
            new_pkt.destruct= av_destruct_packet;
        }
        *pkt= new_pkt;

        bsfc= bsfc->next;
    }

    av_interleaved_write_frame(s, pkt);
}

#define MAX_AUDIO_PACKET_SIZE (128 * 1024)

static void do_audio_out(AVFormatContext *s,
                         AVOutputStream *ost,
                         AVInputStream *ist,
                         unsigned char *buf, int size,
                         Options& options)
{
    uint8_t *buftmp;
    static uint8_t *audio_buf = NULL;
    static uint8_t *audio_out = NULL;
    static uint8_t *audio_out2 = NULL;
    const int audio_out_size= 4*MAX_AUDIO_PACKET_SIZE;

    int size_out, frame_bytes, ret;
    AVCodecContext *enc= ost->st->codec;
    AVCodecContext *dec= ist->st->codec;
    int osize= av_get_bits_per_sample_format(enc->sample_fmt)/8;
    int isize= av_get_bits_per_sample_format(dec->sample_fmt)/8;

    /* SC: dynamic allocation of buffers */
    if (!audio_buf)
        audio_buf = (uint8_t*)av_malloc(2*MAX_AUDIO_PACKET_SIZE);
    if (!audio_out)
        audio_out = (uint8_t*)av_malloc(audio_out_size);
    if (!audio_buf || !audio_out)
        return;               /* Should signal an error ! */

    if (enc->channels != dec->channels)
        ost->audio_resample = 1;

    if (ost->audio_resample && !ost->resample) {
        if (dec->sample_fmt != SAMPLE_FMT_S16)
            fprintf(stderr, "Warning, using s16 intermediate sample format for resampling\n");
        ost->resample = av_audio_resample_init(enc->channels,    dec->channels,
                                               enc->sample_rate, dec->sample_rate,
                                               enc->sample_fmt,  dec->sample_fmt,
                                               16, 10, 0, 0.8);
        if (!ost->resample) {
            fprintf(stderr, "Can not resample %d channels @ %d Hz to %d channels @ %d Hz\n",
                    dec->channels, dec->sample_rate,
                    enc->channels, enc->sample_rate);
            throw std::exception();
        }
    }

#define MAKE_SFMT_PAIR(a,b) ((a)+SAMPLE_FMT_NB*(b))
    if (!ost->audio_resample && dec->sample_fmt!=enc->sample_fmt &&
        MAKE_SFMT_PAIR(enc->sample_fmt,dec->sample_fmt)!=ost->reformat_pair) {
        if (!audio_out2)
            audio_out2 = (uint8_t*)av_malloc(audio_out_size);
        if (!audio_out2)
           throw std::exception();
        if (ost->reformat_ctx)
            av_audio_convert_free(ost->reformat_ctx);
        ost->reformat_ctx = av_audio_convert_alloc(enc->sample_fmt, 1,
                                                   dec->sample_fmt, 1, NULL, 0);
        if (!ost->reformat_ctx) {
            fprintf(stderr, "Cannot convert %s sample format to %s sample format\n",
                avcodec_get_sample_fmt_name(dec->sample_fmt),
                avcodec_get_sample_fmt_name(enc->sample_fmt));
            throw std::exception();
        }
        ost->reformat_pair=MAKE_SFMT_PAIR(enc->sample_fmt,dec->sample_fmt);
    }


    if(options.audio_sync_method){
        double delta = get_sync_ipts(ost, options) * enc->sample_rate - ost->sync_opts
                - av_fifo_size(&ost->fifo)/(ost->st->codec->channels * 2);
        double idelta= delta*ist->st->codec->sample_rate / enc->sample_rate;
        int byte_delta= ((int)idelta)*2*ist->st->codec->channels;

        //FIXME resample delay
        if(fabs(delta) > 50){
            if(ist->is_start || fabs(delta) > options.audio_drift_threshold*enc->sample_rate){
                if(byte_delta < 0){
                    byte_delta= FFMAX(byte_delta, -size);
                    size += byte_delta;
                    buf  -= byte_delta;
                    if(options.verbose > 2)
                        fprintf(stderr, "discarding %d audio samples\n", (int)-delta);
                    if(!size)
                        return;
                    ist->is_start=0;
                }else{
                    static uint8_t *input_tmp= NULL;
                    input_tmp= (uint8_t*)av_realloc(input_tmp, byte_delta + size);

                    if(byte_delta + size <= MAX_AUDIO_PACKET_SIZE)
                        ist->is_start=0;
                    else
                        byte_delta= MAX_AUDIO_PACKET_SIZE - size;

                    memset(input_tmp, 0, byte_delta);
                    memcpy(input_tmp + byte_delta, buf, size);
                    buf= input_tmp;
                    size += byte_delta;
                    if(options.verbose > 2)
                        fprintf(stderr, "adding %d audio samples of silence\n", (int)delta);
                }
            }else if(options.audio_sync_method>1){
                int comp= av_clip(static_cast<int>(delta), -options.audio_sync_method, options.audio_sync_method);
                assert(ost->audio_resample);
                if(options.verbose > 2)
                    fprintf(stderr, "compensating audio timestamp drift:%f compensation:%d in:%d\n", delta, comp, enc->sample_rate);
//                fprintf(stderr, "drift:%f len:%d opts:%"PRId64" ipts:%"PRId64" fifo:%d\n", delta, -1, ost->sync_opts, (int64_t)(get_sync_ipts(ost) * enc->sample_rate), av_fifo_size(&ost->fifo)/(ost->st->codec->channels * 2));
                av_resample_compensate(*(struct AVResampleContext**)ost->resample, comp, enc->sample_rate);
            }
        }
    }else
        ost->sync_opts= lrintf(static_cast<float>(get_sync_ipts(ost, options)) * enc->sample_rate)
                        - av_fifo_size(&ost->fifo)/(ost->st->codec->channels * 2); //FIXME wrong

    if (ost->audio_resample) {
        buftmp = audio_buf;
        size_out = audio_resample(ost->resample,
                                  (short *)buftmp, (short *)buf,
                                  size / (ist->st->codec->channels * isize));
        size_out = size_out * enc->channels * osize;
    } else {
        buftmp = buf;
        size_out = size;
    }

    if (!ost->audio_resample && dec->sample_fmt!=enc->sample_fmt) {
        const void *ibuf[6]= {buftmp};
        void *obuf[6]= {audio_out2};
        int istride[6]= {isize};
        int ostride[6]= {osize};
        int len= size_out/istride[0];
        if (av_audio_convert(ost->reformat_ctx, obuf, ostride, ibuf, istride, len)<0) {
            printf("av_audio_convert() failed\n");
            return;
        }
        buftmp = audio_out2;
        size_out = len*osize;
    }

    /* now encode as many frames as possible */
    if (enc->frame_size > 1) {
        /* output resampled raw samples */
        if (av_fifo_realloc2(&ost->fifo, av_fifo_size(&ost->fifo) + size_out) < 0) {
            fprintf(stderr, "av_fifo_realloc2() failed\n");
            throw std::exception();
        }
        av_fifo_generic_write(&ost->fifo, buftmp, size_out, NULL);

        frame_bytes = enc->frame_size * osize * enc->channels;

        while (av_fifo_size(&ost->fifo) >= frame_bytes) {
            AVPacket pkt;
            av_init_packet(&pkt);

            av_fifo_read(&ost->fifo, audio_buf, frame_bytes);

            //FIXME pass ost->sync_opts as AVFrame.pts in avcodec_encode_audio()
            
            ret = avcodec_encode_audio(enc, audio_out, audio_out_size,
                                       (short *)audio_buf);
            if (ret < 0) {
                fprintf(stderr, "Audio encoding failed\n");
                //av_exit(1);
                throw std::exception();
            }
            options.audio_size += ret;
            pkt.stream_index= ost->index;
            pkt.data= audio_out;
            pkt.size= ret;
            if(enc->coded_frame && enc->coded_frame->pts != AV_NOPTS_VALUE)
                pkt.pts= av_rescale_q(enc->coded_frame->pts, enc->time_base, ost->st->time_base);
            pkt.flags |= PKT_FLAG_KEY;
            write_frame(s, &pkt, ost->st->codec, options.bitstream_filters[ost->file_index][pkt.stream_index]);

            ost->sync_opts += enc->frame_size;
        }
    } else {
        AVPacket pkt;
        int coded_bps = av_get_bits_per_sample(enc->codec->id)/8;
        av_init_packet(&pkt);

        ost->sync_opts += size_out / (osize * enc->channels);

        /* output a pcm frame */
        /* determine the size of the coded buffer */
        size_out /= osize;
        if (coded_bps)
            size_out *= coded_bps;

        //FIXME pass ost->sync_opts as AVFrame.pts in avcodec_encode_audio()
        ret = avcodec_encode_audio(enc, audio_out, size_out,
                                   (short *)buftmp);
        if (ret < 0) {
            fprintf(stderr, "Audio encoding failed\n");
            //av_exit(1);
            throw std::exception();
        }
        options.audio_size += ret;
        pkt.stream_index= ost->index;
        pkt.data= audio_out;
        pkt.size= ret;
        if(enc->coded_frame && enc->coded_frame->pts != AV_NOPTS_VALUE)
            pkt.pts= av_rescale_q(enc->coded_frame->pts, enc->time_base, ost->st->time_base);
        pkt.flags |= PKT_FLAG_KEY;
        write_frame(s, &pkt, ost->st->codec, options.bitstream_filters[ost->file_index][pkt.stream_index]);
    }
}

static void pre_process_video_frame(AVInputStream *ist, AVPicture *picture, void **bufp, Options& options)
{
    AVCodecContext *dec;
    AVPicture *picture2;
    AVPicture picture_tmp;
    uint8_t *buf = 0;

    dec = ist->st->codec;

    /* deinterlace : must be done before any resize */
    if (options.do_deinterlace || options.using_vhook) {
        int size;

        /* create temporary picture */
        size = avpicture_get_size(dec->pix_fmt, dec->width, dec->height);
        buf = (uint8_t*)av_malloc(size);
        if (!buf)
            return;

        picture2 = &picture_tmp;
        avpicture_fill(picture2, buf, dec->pix_fmt, dec->width, dec->height);

        if (options.do_deinterlace){
            if(avpicture_deinterlace(picture2, picture,
                                     dec->pix_fmt, dec->width, dec->height) < 0) {
                /* if error, do not deinterlace */
                av_free(buf);
                buf = NULL;
                picture2 = picture;
            }
        } else {
            av_picture_copy(picture2, picture, dec->pix_fmt, dec->width, dec->height);
        }
    } else {
        picture2 = picture;
    }
/*
    if (ENABLE_VHOOK)
        frame_hook_process(picture2, dec->pix_fmt, dec->width, dec->height,
                           1000000 * ist->pts / AV_TIME_BASE);
*/

    if (picture != picture2)
        *picture = *picture2;
    *bufp = buf;
}

/* we begin to correct av delay at this threshold */
#define AV_DELAY_MAX 0.100

static void do_subtitle_out(AVFormatContext *s,
                            AVOutputStream *ost,
                            AVInputStream *ist,
                            AVSubtitle *sub,
                            int64_t pts,
                            Options& options)
{
    static uint8_t *subtitle_out = NULL;
    int subtitle_out_max_size = 65536;
    int subtitle_out_size, nb, i;
    AVCodecContext *enc;
    AVPacket pkt;

    if (pts == AV_NOPTS_VALUE) {
        fprintf(stderr, "Subtitle packets must have a pts\n");
        return;
    }

    enc = ost->st->codec;

    if (!subtitle_out) {
        subtitle_out = (uint8_t*)av_malloc(subtitle_out_max_size);
    }

    /* Note: DVB subtitle need one packet to draw them and one other
       packet to clear them */
    /* XXX: signal it in the codec context ? */
    if (enc->codec_id == CODEC_ID_DVB_SUBTITLE)
        nb = 2;
    else
        nb = 1;

    for(i = 0; i < nb; i++) {
        subtitle_out_size = avcodec_encode_subtitle(enc, subtitle_out,
                                                    subtitle_out_max_size, sub);

        av_init_packet(&pkt);
        pkt.stream_index = ost->index;
        pkt.data = subtitle_out;
        pkt.size = subtitle_out_size;
        pkt.pts = av_rescale_q(pts, ist->st->time_base, ost->st->time_base);
        if (enc->codec_id == CODEC_ID_DVB_SUBTITLE) {
            /* XXX: the pts correction is handled here. Maybe handling
               it in the codec would be better */
            if (i == 0)
                pkt.pts += 90 * sub->start_display_time;
            else
                pkt.pts += 90 * sub->end_display_time;
        }
        write_frame(s, &pkt, ost->st->codec, options.bitstream_filters[ost->file_index][pkt.stream_index]);
    }
}

static int bit_buffer_size= 1024*256;
static uint8_t *bit_buffer= NULL;

static void do_video_out(AVFormatContext *s,
                         AVOutputStream *ost,
                         AVInputStream *ist,
                         AVFrame *in_picture,
                         int *frame_size,
                         Options& options)
{
    int nb_frames, i, ret;
    AVFrame *final_picture, *formatted_picture, *resampling_dst, *padding_src;
    AVFrame picture_crop_temp, picture_pad_temp;
    AVCodecContext *enc, *dec;

    avcodec_get_frame_defaults(&picture_crop_temp);
    avcodec_get_frame_defaults(&picture_pad_temp);

    enc = ost->st->codec;
    dec = ist->st->codec;

    /* by default, we output a single frame */
    nb_frames = 1;

    *frame_size = 0;

    if(options.video_sync_method>0 || (options.video_sync_method && av_q2d(enc->time_base) > 0.001)){
        double vdelta;
        vdelta = get_sync_ipts(ost, options) / av_q2d(enc->time_base) - ost->sync_opts;
        //FIXME set to 0.5 after we fix some dts/pts bugs like in avidec.c
        if (vdelta < -1.1)
            nb_frames = 0;
        else if (options.video_sync_method == 2)
            ost->sync_opts= lrintf(get_sync_ipts(ost, options) / av_q2d(enc->time_base));
        else if (vdelta > 1.1)
            nb_frames = lrintf(static_cast<float>(vdelta));
//fprintf(stderr, "vdelta:%f, ost->sync_opts:%"PRId64", ost->sync_ipts:%f nb_frames:%d\n", vdelta, ost->sync_opts, ost->sync_ipts, nb_frames);
        if (nb_frames == 0){
            ++options.nb_frames_drop;
            if (options.verbose>2)
                fprintf(stderr, "*** drop!\n");
        }else if (nb_frames > 1) {
            options.nb_frames_dup += nb_frames;
            if (options.verbose>2)
                fprintf(stderr, "*** %d dup!\n", nb_frames-1);
        }
    }else
        ost->sync_opts= lrintf(static_cast<float>(get_sync_ipts(ost, options) / av_q2d(enc->time_base)));

    nb_frames= FFMIN(nb_frames, options.max_frames[CODEC_TYPE_VIDEO] - ost->frame_number);
    if (nb_frames <= 0)
        return;

    if (ost->video_crop) {
        if (av_picture_crop((AVPicture *)&picture_crop_temp, (AVPicture *)in_picture, dec->pix_fmt, ost->topBand, ost->leftBand) < 0) {
            av_log(NULL, AV_LOG_ERROR, "error cropping picture\n");
            return;
        }
        formatted_picture = &picture_crop_temp;
    } else {
        formatted_picture = in_picture;
    }

    final_picture = formatted_picture;
    padding_src = formatted_picture;
    resampling_dst = &ost->pict_tmp;
    if (ost->video_pad) {
        final_picture = &ost->pict_tmp;
        if (ost->video_resample) {
            if (av_picture_crop((AVPicture *)&picture_pad_temp, (AVPicture *)final_picture, enc->pix_fmt, ost->padtop, ost->padleft) < 0) {
                av_log(NULL, AV_LOG_ERROR, "error padding picture\n");
                return;
            }
            resampling_dst = &picture_pad_temp;
        }
    }

    if (ost->video_resample) {
        padding_src = NULL;
        final_picture = &ost->pict_tmp;
        sws_scale(ost->img_resample_ctx, formatted_picture->data, formatted_picture->linesize,
              0, ost->resample_height, resampling_dst->data, resampling_dst->linesize);
    }

    if (ost->video_pad) {
        av_picture_pad((AVPicture*)final_picture, (AVPicture *)padding_src,
                enc->height, enc->width, enc->pix_fmt,
                ost->padtop, ost->padbottom, ost->padleft, ost->padright, options.padcolor);
    }

    /* duplicates frame if needed */
    for(i=0;i<nb_frames;i++) {
        AVPacket pkt;
        av_init_packet(&pkt);
        pkt.stream_index= ost->index;

        if (s->oformat->flags & AVFMT_RAWPICTURE) {
            /* raw pictures are written as AVPicture structure to
               avoid any copies. We support temorarily the older
               method. */
            AVFrame* old_frame = enc->coded_frame;
            enc->coded_frame = dec->coded_frame; //FIXME/XXX remove this hack
            pkt.data= (uint8_t *)final_picture;
            pkt.size=  sizeof(AVPicture);
            pkt.pts= av_rescale_q(ost->sync_opts, enc->time_base, ost->st->time_base);
            pkt.flags |= PKT_FLAG_KEY;

            write_frame(s, &pkt, ost->st->codec, options.bitstream_filters[ost->file_index][pkt.stream_index]);
            enc->coded_frame = old_frame;
        } else {
            AVFrame big_picture;

            big_picture= *final_picture;
            /* better than nothing: use input picture interlaced
               settings */
            big_picture.interlaced_frame = in_picture->interlaced_frame;
            if(options.avctx_opts[CODEC_TYPE_VIDEO]->flags & (CODEC_FLAG_INTERLACED_DCT|CODEC_FLAG_INTERLACED_ME)){
                if(options.top_field_first == -1)
                    big_picture.top_field_first = in_picture->top_field_first;
                else
                    big_picture.top_field_first = options.top_field_first;
            }

            /* handles sameq here. This is not correct because it may
               not be a global option */
            if (options.same_quality) {
                big_picture.quality = static_cast<int>(ist->st->quality);
            }else
                big_picture.quality = static_cast<int>(ost->st->quality);
            if(!options.me_threshold)
                big_picture.pict_type = 0;
//            big_picture.pts = AV_NOPTS_VALUE;
            big_picture.pts= ost->sync_opts;
//            big_picture.pts= av_rescale(ost->sync_opts, AV_TIME_BASE*(int64_t)enc->time_base.num, enc->time_base.den);
//av_log(NULL, AV_LOG_DEBUG, "%"PRId64" -> encoder\n", ost->sync_opts);
            ret = avcodec_encode_video(enc,
                                       bit_buffer, bit_buffer_size,
                                       &big_picture);
            if (ret == -1) {
                throw std::exception();// std::exception("Video encoding failed\n");
            }
            //enc->frame_number = enc->real_pict_num;
            if(ret>0){
                pkt.data= bit_buffer;
                pkt.size= ret;
                if(enc->coded_frame && enc->coded_frame->pts != AV_NOPTS_VALUE)
                    pkt.pts= av_rescale_q(enc->coded_frame->pts, enc->time_base, ost->st->time_base);
/*av_log(NULL, AV_LOG_DEBUG, "encoder -> %"PRId64"/%"PRId64"\n",
   pkt.pts != AV_NOPTS_VALUE ? av_rescale(pkt.pts, enc->time_base.den, AV_TIME_BASE*(int64_t)enc->time_base.num) : -1,
   pkt.dts != AV_NOPTS_VALUE ? av_rescale(pkt.dts, enc->time_base.den, AV_TIME_BASE*(int64_t)enc->time_base.num) : -1);*/

                if(enc->coded_frame && enc->coded_frame->key_frame)
                    pkt.flags |= PKT_FLAG_KEY;
                write_frame(s, &pkt, ost->st->codec, options.bitstream_filters[ost->file_index][pkt.stream_index]);
                *frame_size = ret;
                options.video_size += ret;
                //fprintf(stderr,"\nFrame: %3d %3d size: %5d type: %d",
                //        enc->frame_number-1, enc->real_pict_num, ret,
                //        enc->pict_type);
                /* if two pass, output log */
                if (ost->logfile && enc->stats_out) {
                    fprintf(ost->logfile, "%s", enc->stats_out);
                }
            }
        }
        ost->sync_opts++;
        ost->frame_number++;
    }
}

static double psnr(double d){
    if(d==0) return INFINITY;
    return -10.0*log(d)/log(10.0);
}

static void do_video_stats(AVFormatContext *os, AVOutputStream *ost,
                           int frame_size, Options& options)
{
    AVCodecContext *enc;
    int frame_number;
    double ti1, bitrate, avg_bitrate;

    /* this is executed just the first time do_video_stats is called */
    if (!options.fvstats) {
        options.fvstats = fopen(options.vstats_filename, "w");
        if (!options.fvstats) {
            throw std::exception(); // std::exception("fopen");
        }
    }

    enc = ost->st->codec;
    if (enc->codec_type == CODEC_TYPE_VIDEO) {
        frame_number = ost->frame_number;
        fprintf(options.fvstats, "frame= %5d q= %2.1f ", frame_number, enc->coded_frame->quality/(float)FF_QP2LAMBDA);
        if (enc->flags&CODEC_FLAG_PSNR)
            fprintf(options.fvstats, "PSNR= %6.2f ", psnr(enc->coded_frame->error[0]/(enc->width*enc->height*255.0*255.0)));

        fprintf(options.fvstats,"f_size= %6d ", frame_size);
        /* compute pts value */
        ti1 = ost->sync_opts * av_q2d(enc->time_base);
        if (ti1 < 0.01)
            ti1 = 0.01;

        bitrate = (frame_size * 8) / av_q2d(enc->time_base) / 1000.0;
        avg_bitrate = (double)(options.video_size * 8) / ti1 / 1000.0;
        fprintf(options.fvstats, "s_size= %8.0fkB time= %0.3f br= %7.1fkbits/s avg_br= %7.1fkbits/s ",
            (double)options.video_size / 1024, ti1, bitrate, avg_bitrate);
        fprintf(options.fvstats,"type= %c\n", av_get_pict_type_char(enc->coded_frame->pict_type));
    }
}

static void print_report(AVFormatContext **output_files,
                         AVOutputStream **ost_table, int nb_ostreams,
                         int is_last_report, Options& options)
{
    char buf[1024];
    AVOutputStream *ost;
    AVFormatContext *oc, *os;
    int64_t total_size;
    AVCodecContext *enc;
    int frame_number, vid, i;
    double bitrate, ti1, pts;
    static int64_t last_time = -1;
    static int qp_histogram[52];

    if (!is_last_report) {
        int64_t cur_time;
        /* display the report every 0.5 seconds */
        cur_time = av_gettime();
        if (last_time == -1) {
            last_time = cur_time;
            return;
        }
        if ((cur_time - last_time) < 500000)
            return;
        last_time = cur_time;
    }


    oc = output_files[0];

    total_size = url_fsize(oc->pb);
    if(total_size<0) // FIXME improve url_fsize() so it works with non seekable output too
        total_size= url_ftell(oc->pb);

    buf[0] = '\0';
    ti1 = 1e10;
    vid = 0;
    for(i=0;i<nb_ostreams;i++) {
        ost = ost_table[i];
        os = output_files[ost->file_index];
        enc = ost->st->codec;
        if (vid && enc->codec_type == CODEC_TYPE_VIDEO) {
            snprintf(buf + strlen(buf), sizeof(buf) - strlen(buf), "q=%2.1f ",
                    !ost->st->stream_copy ?
                    enc->coded_frame->quality/(float)FF_QP2LAMBDA : -1);
        }
        if (!vid && enc->codec_type == CODEC_TYPE_VIDEO) {
            float t = static_cast<float>((av_gettime()-options.timer_start) / 1000000.0);

            frame_number = ost->frame_number;
            snprintf(buf + strlen(buf), sizeof(buf) - strlen(buf), "frame=%5d fps=%3d q=%3.1f ",
                     frame_number, (t>1)?(int)(frame_number/t+0.5) : 0,
                     !ost->st->stream_copy ?
                     enc->coded_frame->quality/(float)FF_QP2LAMBDA : -1);
            if(is_last_report)
                snprintf(buf + strlen(buf), sizeof(buf) - strlen(buf), "L");
            if(options.qp_hist){
                int j;
                int qp= lrintf(enc->coded_frame->quality/(float)FF_QP2LAMBDA);
                if(qp>=0 && qp<sizeof(qp_histogram)/sizeof(int))
                    qp_histogram[qp]++;
                for(j=0; j<32; j++)
                    snprintf(buf + strlen(buf), sizeof(buf) - strlen(buf), "%X", (int)lrintf(static_cast<float>(log((double)(qp_histogram[j]+1))/log(2.0))));
            }
            if (enc->flags&CODEC_FLAG_PSNR){
                int j;
                double error, error_sum=0;
                double scale, scale_sum=0;
                char type[3]= {'Y','U','V'};
                snprintf(buf + strlen(buf), sizeof(buf) - strlen(buf), "PSNR=");
                for(j=0; j<3; j++){
                    if(is_last_report){
                        error= static_cast<double>(enc->error[j]);
                        scale= enc->width*enc->height*255.0*255.0*frame_number;
                    }else{
                        error= static_cast<double>(enc->coded_frame->error[j]);
                        scale= enc->width*enc->height*255.0*255.0;
                    }
                    if(j) scale/=4;
                    error_sum += error;
                    scale_sum += scale;
                    snprintf(buf + strlen(buf), sizeof(buf) - strlen(buf), "%c:%2.2f ", type[j], psnr(error/scale));
                }
                snprintf(buf + strlen(buf), sizeof(buf) - strlen(buf), "*:%2.2f ", psnr(error_sum/scale_sum));
            }
            vid = 1;
        }
        /* compute min output value */
        pts = (double)ost->st->pts.val * av_q2d(ost->st->time_base);
        if ((pts < ti1) && (pts > 0))
            ti1 = pts;
    }
    if (ti1 < 0.01)
        ti1 = 0.01;

    if (options.verbose || is_last_report) {
        bitrate = (double)(total_size * 8) / ti1 / 1000.0;

        snprintf(buf + strlen(buf), sizeof(buf) - strlen(buf),
            "size=%8.0fkB time=%0.2f bitrate=%6.1fkbits/s",
            (double)total_size / 1024, ti1, bitrate);

        if (options.verbose > 1)
          snprintf(buf + strlen(buf), sizeof(buf) - strlen(buf), " dup=%d drop=%d",
                  options.nb_frames_dup, options.nb_frames_drop);

        if (options.verbose >= 0)
            fprintf(stderr, "%s    \r", buf);

        fflush(stderr);
    }

    if (is_last_report && options.verbose >= 0){
        int64_t raw= options.audio_size + options.video_size + options.extra_size;
        fprintf(stderr, "\n");
        fprintf(stderr, "video:%1.0fkB audio:%1.0fkB global headers:%1.0fkB muxing overhead %f%%\n",
                options.video_size/1024.0,
                options.audio_size/1024.0,
                options.extra_size/1024.0,
                100.0*(total_size - raw)/raw
        );
    }
}

/* pkt = NULL means EOF (needed to flush decoder buffers) */
static int output_packet(AVInputStream *ist, int ist_index,
                         AVOutputStream **ost_table, int nb_ostreams,
                         const AVPacket *pkt, Options& options)
{
    AVFormatContext *os;
    AVOutputStream *ost;
    uint8_t *ptr;
    int len, ret, i;
    uint8_t *data_buf;
    int data_size, got_picture;
    AVFrame picture;
    void *buffer_to_free;
    static unsigned int samples_size= 0;
    static short *samples= NULL;
    AVSubtitle subtitle, *subtitle_to_free;
    int got_subtitle;

    if(ist->next_pts == AV_NOPTS_VALUE)
        ist->next_pts= ist->pts;

    if (pkt == NULL) {
        /* EOF handling */
        ptr = NULL;
        len = 0;
        goto handle_eof;
    }

    if(pkt->dts != AV_NOPTS_VALUE)
    {
		AVRational tbq = AV_TIME_BASE_Q;
        ist->next_pts = ist->pts = av_rescale_q(pkt->dts, ist->st->time_base, tbq);
    }

    len = pkt->size;
    ptr = pkt->data;

    //while we have more to decode or while the decoder did output something on EOF
    while (len > 0 || (!pkt && ist->next_pts != ist->pts)) {
    handle_eof:
		ist->pts= ist->next_pts;

        /* decode the packet if needed */
        data_buf = NULL; /* fail safe */
        data_size = 0;
        subtitle_to_free = NULL;
        if (ist->decoding_needed) {
            switch(ist->st->codec->codec_type) {
            case CODEC_TYPE_AUDIO:{
                if(pkt && samples_size < FFMAX(pkt->size*sizeof(*samples), AVCODEC_MAX_AUDIO_FRAME_SIZE)) {
                    samples_size = FFMAX(pkt->size*sizeof(*samples), AVCODEC_MAX_AUDIO_FRAME_SIZE);
                    av_free(samples);
                    samples= (short int*)av_malloc(samples_size);
                }
                data_size= samples_size;
                    /* XXX: could avoid copy if PCM 16 bits with same
                       endianness as CPU */
                ret = avcodec_decode_audio2(ist->st->codec, samples, &data_size,
                                           ptr, len);
                if (ret < 0)
                    goto fail_decode;
                ptr += ret;
                len -= ret;
                /* Some bug in mpeg audio decoder gives */
                /* data_size < 0, it seems they are overflows */
                if (data_size <= 0) {
                    /* no audio frame */
                    continue;
                }
                data_buf = (uint8_t *)samples;
                ist->next_pts += ((int64_t)AV_TIME_BASE/2 * data_size) /
                    (ist->st->codec->sample_rate * ist->st->codec->channels);
                break;}
            case CODEC_TYPE_VIDEO:
                    data_size = (ist->st->codec->width * ist->st->codec->height * 3) / 2;
                    /* XXX: allocate picture correctly */
                    avcodec_get_frame_defaults(&picture);

                    ret = avcodec_decode_video(ist->st->codec,
                                               &picture, &got_picture, ptr, len);
                    ist->st->quality= static_cast<float>(picture.quality);
                    if (ret < 0)
                        goto fail_decode;
                    if (!got_picture) {
                        /* no picture yet */
                        goto discard_packet;
                    }
                    if (ist->st->codec->time_base.num != 0) {
                        ist->next_pts += ((int64_t)AV_TIME_BASE *
                                          ist->st->codec->time_base.num) /
                            ist->st->codec->time_base.den;
                    }
                    len = 0;
                    break;
            case CODEC_TYPE_SUBTITLE:
                ret = avcodec_decode_subtitle(ist->st->codec,
                                              &subtitle, &got_subtitle, ptr, len);
                if (ret < 0)
                    goto fail_decode;
                if (!got_subtitle) {
                    goto discard_packet;
                }
                subtitle_to_free = &subtitle;
                len = 0;
                break;
            default:
                goto fail_decode;
            }
        } else {
            switch(ist->st->codec->codec_type) {
            case CODEC_TYPE_AUDIO:
                ist->next_pts += ((int64_t)AV_TIME_BASE * ist->st->codec->frame_size) /
                    ist->st->codec->sample_rate;
                break;
            case CODEC_TYPE_VIDEO:
                if (ist->st->codec->time_base.num != 0) {
                    ist->next_pts += ((int64_t)AV_TIME_BASE *
                                      ist->st->codec->time_base.num) /
                        ist->st->codec->time_base.den;
                }
                break;
            }
            data_buf = ptr;
            data_size = len;
            ret = len;
            len = 0;
        }

        buffer_to_free = NULL;
        if (ist->st->codec->codec_type == CODEC_TYPE_VIDEO) {
            pre_process_video_frame(ist, (AVPicture *)&picture,
                                    &buffer_to_free, options);
        }

        // preprocess audio (volume)
        if (ist->st->codec->codec_type == CODEC_TYPE_AUDIO) {
            if (options.audio_volume != 256) {
                short *volp;
                volp = samples;
                for(i=0;i<static_cast<int>(data_size / sizeof(short));i++) {
                    int v = ((*volp) * options.audio_volume + 128) >> 8;
                    if (v < -32768) v = -32768;
                    if (v >  32767) v = 32767;
                    *volp++ = v;
                }
            }
        }

        /* frame rate emulation */
        if (ist->st->codec->rate_emu) {
            int64_t pts = av_rescale((int64_t) ist->frame * ist->st->codec->time_base.num, 1000000, ist->st->codec->time_base.den);
            int64_t now = av_gettime() - ist->start;
            if (pts > now)
#ifdef _WIN32
                Sleep(static_cast<DWORD>((pts - now) * 10));
#else
                usleep(pts - now);
#endif

            ist->frame++;
        }

#if 0
        /* mpeg PTS deordering : if it is a P or I frame, the PTS
           is the one of the next displayed one */
        /* XXX: add mpeg4 too ? */
        if (ist->st->codec->codec_id == CODEC_ID_MPEG1VIDEO) {
            if (ist->st->codec->pict_type != B_TYPE) {
                int64_t tmp;
                tmp = ist->last_ip_pts;
                ist->last_ip_pts  = ist->frac_pts.val;
                ist->frac_pts.val = tmp;
            }
        }
#endif
        /* if output time reached then transcode raw format,
           encode packets and output them */
        if (options.start_time == 0 || ist->pts >= options.start_time)
            for(i=0;i<nb_ostreams;i++) {
                int frame_size;

                ost = ost_table[i];
                if (ost->source_index == ist_index) {
                    os = options.output_files[ost->file_index];

#if 0
                    printf("%d: got pts=%0.3f %0.3f\n", i,
                           (double)pkt->pts / AV_TIME_BASE,
                           ((double)ist->pts / AV_TIME_BASE) -
                           ((double)ost->st->pts.val * ost->st->time_base.num / ost->st->time_base.den));
#endif
                    /* set the input output pts pairs */
                    //ost->sync_ipts = (double)(ist->pts + input_files_ts_offset[ist->file_index] - options.start_time)/ AV_TIME_BASE;

                    if (ost->encoding_needed) {
                        switch(ost->st->codec->codec_type) {
                        case CODEC_TYPE_AUDIO:
                            do_audio_out(os, ost, ist, data_buf, data_size, options);
                            break;
                        case CODEC_TYPE_VIDEO:
                            do_video_out(os, ost, ist, &picture, &frame_size, options);
                            if (options.vstats_filename && frame_size)
                                do_video_stats(os, ost, frame_size, options);
                            break;
                        case CODEC_TYPE_SUBTITLE:
                            do_subtitle_out(os, ost, ist, &subtitle,
                                            pkt->pts, options);
                            break;
                        default:
                            ThrowInternalError();
                        }
                    } else {
                        AVFrame avframe; //FIXME/XXX remove this
                        AVPacket opkt;
                        av_init_packet(&opkt);

                        if (!ost->frame_number && !(pkt->flags & PKT_FLAG_KEY))
                            continue;

                        /* no reencoding needed : output the packet directly */
                        /* force the input stream PTS */

                        avcodec_get_frame_defaults(&avframe);
                        ost->st->codec->coded_frame= &avframe;
                        avframe.key_frame = pkt->flags & PKT_FLAG_KEY;

                        if(ost->st->codec->codec_type == CODEC_TYPE_AUDIO)
                            options.audio_size += data_size;
                        else if (ost->st->codec->codec_type == CODEC_TYPE_VIDEO) {
                            options.video_size += data_size;
                            ost->sync_opts++;
                        }

                        opkt.stream_index= ost->index;
                        if(pkt->pts != AV_NOPTS_VALUE)
                            opkt.pts= av_rescale_q(pkt->pts, ist->st->time_base, ost->st->time_base);
                        else
                            opkt.pts= AV_NOPTS_VALUE;

                        if (pkt->dts == AV_NOPTS_VALUE)
                        {
                            AVRational tbq = AV_TIME_BASE_Q;
                            opkt.dts = av_rescale_q(ist->pts, tbq, ost->st->time_base);
                        }
                        else
                            opkt.dts = av_rescale_q(pkt->dts, ist->st->time_base, ost->st->time_base);

                        opkt.duration = static_cast<int>(av_rescale_q(pkt->duration, ist->st->time_base, ost->st->time_base));
                        opkt.flags= pkt->flags;

                        //FIXME remove the following 2 lines they shall be replaced by the bitstream filters
                        if(av_parser_change(ist->st->parser, ost->st->codec, &opkt.data, &opkt.size, data_buf, data_size, pkt->flags & PKT_FLAG_KEY))
                            opkt.destruct= av_destruct_packet;

                        write_frame(os, &opkt, ost->st->codec, options.bitstream_filters[ost->file_index][pkt->stream_index]);
                        ost->st->codec->frame_number++;
                        ost->frame_number++;
                        av_free_packet(&opkt);
                    }
                }
            }
        av_free(buffer_to_free);
        /* XXX: allocate the subtitles in the codec ? */
        if (subtitle_to_free) {
            if (subtitle_to_free->rects != NULL) {
                for (i = 0; i < static_cast<int>(subtitle_to_free->num_rects); i++) {
                    av_freep(&subtitle_to_free->rects[i]->pict.data[0]);
                    av_freep(&subtitle_to_free->rects[i]->pict.data[1]);
                    av_freep(&subtitle_to_free->rects[i]);
                }
                av_freep(&subtitle_to_free->rects);
            }
            subtitle_to_free->num_rects = 0;
            subtitle_to_free = NULL;
        }
    }
 discard_packet:
    if (pkt == NULL) {
        /* EOF handling */

        for(i=0;i<nb_ostreams;i++) {
            ost = ost_table[i];
            if (ost->source_index == ist_index) {
                AVCodecContext *enc= ost->st->codec;
                os = options.output_files[ost->file_index];

                if(ost->st->codec->codec_type == CODEC_TYPE_AUDIO && enc->frame_size <=1)
                    continue;
                if(ost->st->codec->codec_type == CODEC_TYPE_VIDEO && (os->oformat->flags & AVFMT_RAWPICTURE))
                    continue;

                if (ost->encoding_needed) {
                    for(;;) {
                        AVPacket pkt;
                        int fifo_bytes;
                        av_init_packet(&pkt);
                        pkt.stream_index= ost->index;

                        switch(ost->st->codec->codec_type) {
                        case CODEC_TYPE_AUDIO:
                            fifo_bytes = av_fifo_size(&ost->fifo);
                            ret = 0;
                            /* encode any samples remaining in fifo */
                            if(fifo_bytes > 0 && enc->codec->capabilities & CODEC_CAP_SMALL_LAST_FRAME) {
                                int fs_tmp = enc->frame_size;
                                enc->frame_size = fifo_bytes / (2 * enc->channels);
                                av_fifo_read(&ost->fifo, (uint8_t *)samples, fifo_bytes);
                                    ret = avcodec_encode_audio(enc, bit_buffer, bit_buffer_size, samples);
                                enc->frame_size = fs_tmp;
                            }
                            if(ret <= 0) {
                                ret = avcodec_encode_audio(enc, bit_buffer, bit_buffer_size, NULL);
                            }
                            options.audio_size += ret;
                            pkt.flags |= PKT_FLAG_KEY;
                            break;
                        case CODEC_TYPE_VIDEO:
                            ret = avcodec_encode_video(enc, bit_buffer, bit_buffer_size, NULL);
                            options.video_size += ret;
                            if(enc->coded_frame && enc->coded_frame->key_frame)
                                pkt.flags |= PKT_FLAG_KEY;
                            if (ost->logfile && enc->stats_out) {
                                fprintf(ost->logfile, "%s", enc->stats_out);
                            }
                            break;
                        default:
                            ret=-1;
                        }

                        if(ret<=0)
                            break;
                        pkt.data= bit_buffer;
                        pkt.size= ret;
                        if(enc->coded_frame && enc->coded_frame->pts != AV_NOPTS_VALUE)
                            pkt.pts= av_rescale_q(enc->coded_frame->pts, enc->time_base, ost->st->time_base);
                        write_frame(os, &pkt, ost->st->codec, options.bitstream_filters[ost->file_index][pkt.stream_index]);
                    }
                }
            }
        }
    }

    return 0;
 fail_decode:
    return -1;
}

static int stream_index_from_inputs(AVFormatContext **input_files,
                                    int nb_input_files,
                                    AVInputFile *file_table,
                                    AVInputStream **ist_table,
                                    enum CodecType type,
                                    int programid)
{
    int p, q, z;
    for(z=0; z<nb_input_files; z++) {
        AVFormatContext *ic = input_files[z];
        for(p=0; p<ic->nb_programs; p++) {
            AVProgram *program = ic->programs[p];
            if(program->id != programid)
                continue;
            for(q=0; q<program->nb_stream_indexes; q++) {
                int sidx = program->stream_index[q];
                int ris = file_table[z].ist_index + sidx;
                if(ist_table[ris]->discard && ic->streams[sidx]->codec->codec_type == type)
                    return ris;
            }
        }
    }

    return -1;
}

/*
 * The following code is the main loop of the file converter
 */
static int av_encode(AVFormatContext **output_files,
                     int nb_output_files,
                     AVFormatContext **input_files,
                     int nb_input_files,
                     AVStreamMap *stream_maps, int nb_stream_maps,
                     Options& options,
                     Progress& progress)
{
    int ret, i, j, k, n, nb_istreams = 0, nb_ostreams = 0;
    AVFormatContext *is, *os;
    AVCodecContext *codec, *icodec;
    AVOutputStream *ost, **ost_table = NULL;
    AVInputStream *ist, **ist_table = NULL;
    AVInputFile *file_table;
    int key;
   	const AVRational tbq = AV_TIME_BASE_Q;
	uint64_t totalBytesToProcess = 0;
	float lastPercentComplete = 0.0;

    file_table= (AVInputFile*) av_mallocz(nb_input_files * sizeof(AVInputFile));
    if (!file_table)
        goto fail;

    /* input stream init */
    j = 0;
    for(i=0;i<nb_input_files;i++) {
        is = input_files[i];
        file_table[i].ist_index = j;
        file_table[i].nb_streams = is->nb_streams;
        j += is->nb_streams;

        totalBytesToProcess += url_fsize(is->pb);
    }
    nb_istreams = j;

    ist_table = (AVInputStream**)av_mallocz(nb_istreams * sizeof(AVInputStream *));
    if (!ist_table)
        goto fail;

    for(i=0;i<nb_istreams;i++) {
        ist = (AVInputStream*)av_mallocz(sizeof(AVInputStream));
        if (!ist)
            goto fail;
        ist_table[i] = ist;
    }
    j = 0;
    for(i=0;i<nb_input_files;i++) {
        is = input_files[i];
        for(k=0;k<static_cast<int>(is->nb_streams);k++) {
            ist = ist_table[j++];
            ist->st = is->streams[k];
            ist->file_index = i;
            ist->index = k;
            ist->discard = 1; /* the stream is discarded by default
                                 (changed later) */

            if (ist->st->codec->rate_emu) {
                ist->start = av_gettime();
                ist->frame = 0;
            }
        }
    }

    /* output stream init */
    nb_ostreams = 0;
    for(i=0;i<nb_output_files;i++) {
        os = output_files[i];
        if (!os->nb_streams) {
			throw std::exception();// std::exception("Output file does not contain any stream");
        }
        nb_ostreams += os->nb_streams;
    }
    if (nb_stream_maps > 0 && nb_stream_maps != nb_ostreams) {
		throw std::exception(); // std::exception("Number of stream maps must match number of output streams");
    }

    /* Sanity check the mapping args -- do the input files & streams exist? */
    for(i=0;i<nb_stream_maps;i++) {
        int fi = stream_maps[i].file_index;
        int si = stream_maps[i].stream_index;

        if (fi < 0 || fi > nb_input_files - 1 ||
            si < 0 || si > file_table[fi].nb_streams - 1) {
            throw std::exception(); // std::exception("Could not find input stream");
        }
        fi = stream_maps[i].sync_file_index;
        si = stream_maps[i].sync_stream_index;
        if (fi < 0 || fi > nb_input_files - 1 ||
            si < 0 || si > file_table[fi].nb_streams - 1) {
            throw std::exception(); // std::exception("Could not find sync stream");
        }
    }

    ost_table = (AVOutputStream**)av_mallocz(sizeof(AVOutputStream *) * nb_ostreams);
    if (!ost_table)
        goto fail;
    for(i=0;i<nb_ostreams;i++) {
        ost = (AVOutputStream*)av_mallocz(sizeof(AVOutputStream));
        if (!ost)
            goto fail;
        ost_table[i] = ost;
    }

    n = 0;
    for(k=0;k<nb_output_files;k++) {
        os = output_files[k];
        for(i=0;i<static_cast<int>(os->nb_streams);i++) {
            int found;
            ost = ost_table[n++];
            ost->file_index = k;
            ost->index = i;
            ost->st = os->streams[i];
            if (nb_stream_maps > 0) {
                ost->source_index = file_table[stream_maps[n-1].file_index].ist_index +
                    stream_maps[n-1].stream_index;

                /* Sanity check that the stream types match */
                if (ist_table[ost->source_index]->st->codec->codec_type != ost->st->codec->codec_type) {
 //                   fprintf(stderr, "Codec type mismatch for mapping #%d.%d -> #%d.%d\n",
 //                       stream_maps[n-1].file_index, stream_maps[n-1].stream_index,
 //                       ost->file_index, ost->index);
 //                   exit(1);
                    int i= ost->file_index;
                    dump_format(output_files[i], i, output_files[i]->filename, 1);
                    throw std::exception(); // std::exception("Codec type mismatch");
                }

            } else {
                if(options.opt_programid) {
                    found = 0;
                    j = stream_index_from_inputs(input_files, nb_input_files, file_table, ist_table, ost->st->codec->codec_type, options.opt_programid);
                    if(j != -1) {
                        ost->source_index = j;
                        found = 1;
                    }
                } else {
                    /* get corresponding input stream index : we select the first one with the right type */
                    found = 0;
                    for(j=0;j<nb_istreams;j++) {
                        ist = ist_table[j];
                        if (ist->discard &&
                            ist->st->codec->codec_type == ost->st->codec->codec_type) {
                            ost->source_index = j;
                            found = 1;
                            break;
                        }
                    }
                }

                if (!found) {
                    if(! options.opt_programid) {
                        /* try again and reuse existing stream */
                        for(j=0;j<nb_istreams;j++) {
                            ist = ist_table[j];
                            if (ist->st->codec->codec_type == ost->st->codec->codec_type) {
                                ost->source_index = j;
                                found = 1;
                            }
                        }
                    }
                    if (!found) {
//                        fprintf(stderr, "Could not find input stream matching output stream #%d.%d\n",
//                                ost->file_index, ost->index);
//                        exit(1);
                        int i= ost->file_index;
                        dump_format(output_files[i], i, output_files[i]->filename, 1);
                        throw std::exception(); // std::exception("Could not find input stream matching output stream");
                    }
                }
            }
            ist = ist_table[ost->source_index];
            ist->discard = 0;
            ost->sync_ist = (nb_stream_maps > 0) ?
                ist_table[file_table[stream_maps[n-1].sync_file_index].ist_index +
                         stream_maps[n-1].sync_stream_index] : ist;
        }
    }

    /* for each output stream, we compute the right encoding parameters */
    for(i=0;i<nb_ostreams;i++) {
        ost = ost_table[i];
        os = output_files[ost->file_index];
        ist = ist_table[ost->source_index];

        codec = ost->st->codec;
        icodec = ist->st->codec;

        if (!ost->st->language[0])
            av_strlcpy(ost->st->language, ist->st->language,
                       sizeof(ost->st->language));

        if (ost->st->stream_copy) {
            /* if stream_copy is selected, no need to decode or encode */
            codec->codec_id = icodec->codec_id;
            codec->codec_type = icodec->codec_type;

            if(!codec->codec_tag){
                if(   !os->oformat->codec_tag
                   || av_codec_get_id (os->oformat->codec_tag, icodec->codec_tag) > 0
                   || av_codec_get_tag(os->oformat->codec_tag, icodec->codec_id) <= 0)
                    codec->codec_tag = icodec->codec_tag;
            }

            codec->bit_rate = icodec->bit_rate;
            codec->extradata= icodec->extradata;
            codec->extradata_size= icodec->extradata_size;
            if(av_q2d(icodec->time_base) > av_q2d(ist->st->time_base) && av_q2d(ist->st->time_base) < 1.0/1000)
                codec->time_base = icodec->time_base;
            else
                codec->time_base = ist->st->time_base;
            switch(codec->codec_type) {
            case CODEC_TYPE_AUDIO:
                if(options.audio_volume != 256) {
                    fprintf(stderr,"-acodec copy and -vol are incompatible (frames are not decoded)\n");
                    throw std::exception();
                }
                codec->sample_rate = icodec->sample_rate;
                codec->channels = icodec->channels;
                codec->frame_size = icodec->frame_size;
                codec->block_align= icodec->block_align;
                if(codec->block_align == 1 && codec->codec_id == CODEC_ID_MP3)
                    codec->block_align= 0;
                if(codec->codec_id == CODEC_ID_AC3)
                    codec->block_align= 0;
                break;
            case CODEC_TYPE_VIDEO:
                if(options.using_vhook) {
 //                   fprintf(stderr,"-vcodec copy and -vhook are incompatible (frames are not decoded)\n");
 //                   exit(1);
                    throw std::exception(); // std::exception("-vcodec copy and -vhook are incompatible (frames are not decoded)");
                }
                codec->pix_fmt = icodec->pix_fmt;
                codec->width = icodec->width;
                codec->height = icodec->height;
                codec->has_b_frames = icodec->has_b_frames;
                break;
            case CODEC_TYPE_SUBTITLE:
                break;
            default:
                ThrowInternalError();
            }
        } else {
            switch(codec->codec_type) {
            case CODEC_TYPE_AUDIO:
                if (av_fifo_init(&ost->fifo, 1024))
                    goto fail;
                ost->reformat_pair = MAKE_SFMT_PAIR(SAMPLE_FMT_NONE,SAMPLE_FMT_NONE);
                ost->audio_resample = codec->sample_rate != icodec->sample_rate || options.audio_sync_method > 1;
                icodec->request_channels = codec->channels;
                ist->decoding_needed = 1;
                ost->encoding_needed = 1;
                break;
            case CODEC_TYPE_VIDEO:
                ost->video_crop = ((options.frame_leftBand + options.frame_rightBand + options.frame_topBand + options.frame_bottomBand) != 0);
                ost->video_pad = ((options.frame_padleft + options.frame_padright + options.frame_padtop + options.frame_padbottom) != 0);
                ost->video_resample = ((codec->width != icodec->width -
                                (options.frame_leftBand + options.frame_rightBand) +
                                (options.frame_padleft + options.frame_padright)) ||
                        (codec->height != icodec->height -
                                (options.frame_topBand  + options.frame_bottomBand) +
                                (options.frame_padtop + options.frame_padbottom)) ||
                        (codec->pix_fmt != icodec->pix_fmt));
                if (ost->video_crop) {
                    ost->topBand = options.frame_topBand;
                    ost->leftBand = options.frame_leftBand;
                }
                if (ost->video_pad) {
                    ost->padtop = options.frame_padtop;
                    ost->padleft = options.frame_padleft;
                    ost->padbottom = options.frame_padbottom;
                    ost->padright = options.frame_padright;
                    if (!ost->video_resample) {
                        avcodec_get_frame_defaults(&ost->pict_tmp);
                        if(avpicture_alloc((AVPicture*)&ost->pict_tmp, codec->pix_fmt,
                                         codec->width, codec->height))
                            goto fail;
                    }
                }
                if (ost->video_resample) {
                    avcodec_get_frame_defaults(&ost->pict_tmp);
                    if(avpicture_alloc((AVPicture*)&ost->pict_tmp, codec->pix_fmt,
                                         codec->width, codec->height)) {
//                        fprintf(stderr, "Cannot allocate temp picture, check pix fmt\n");
//                        exit(1);
                        throw std::exception(); // std::exception("Cannot allocate temp picture, check pix fmt");
                    }
                    options.sws_flags = static_cast<int>(av_get_int(options.sws_opts, "sws_flags", NULL));
                    ost->img_resample_ctx = sws_getContext(
                            icodec->width - (options.frame_leftBand + options.frame_rightBand),
                            icodec->height - (options.frame_topBand + options.frame_bottomBand),
                            icodec->pix_fmt,
                            codec->width - (options.frame_padleft + options.frame_padright),
                            codec->height - (options.frame_padtop + options.frame_padbottom),
                            codec->pix_fmt,
                            options.sws_flags, NULL, NULL, NULL);
                    if (ost->img_resample_ctx == NULL) {
 //                       fprintf(stderr, "Cannot get resampling context\n");
 //                       exit(1);
                        throw std::exception(); // std::exception("Cannot get resampling context");
                    }
                    ost->resample_height = icodec->height - (options.frame_topBand + options.frame_bottomBand);
                }
                ost->encoding_needed = 1;
                ist->decoding_needed = 1;
                break;
            case CODEC_TYPE_SUBTITLE:
                ost->encoding_needed = 1;
                ist->decoding_needed = 1;
                break;
            default:
                ThrowInternalError();
                break;
            }
            /* two pass mode */
            if (ost->encoding_needed &&
                (codec->flags & (CODEC_FLAG_PASS1 | CODEC_FLAG_PASS2))) {
                char logfilename[1024];
                FILE *f;
                int size;
                char *logbuffer;

                snprintf(logfilename, sizeof(logfilename), "%s-%d.log",
                         options.pass_logfilename ?
                         options.pass_logfilename : DEFAULT_PASS_LOGFILENAME, i);
                if (codec->flags & CODEC_FLAG_PASS1) {
                    f = fopen(logfilename, "w");
                    if (!f) {
 //                       perror(logfilename);
 //                       exit(1);
                        throw std::exception(); // std::exception("log file could not be opened for write");
                    }
                    ost->logfile = f;
                } else {
                    /* read the log file */
                    f = fopen(logfilename, "r");
                    if (!f) {
 //                       perror(logfilename);
 //                       exit(1);
                        throw std::exception(); // std::exception("log file could not be opened for read");
                    }
                    fseek(f, 0, SEEK_END);
                    size = ftell(f);
                    fseek(f, 0, SEEK_SET);
                    logbuffer = (char*)av_malloc(size + 1);
                    if (!logbuffer) {
//                        fprintf(stderr, "Could not allocate log buffer\n");
//                        exit(1);
                        throw std::exception(); // std::exception("Could not allocate log buffer");
                    }
                    size = static_cast<int>(fread(logbuffer, 1, size, f));
                    fclose(f);
                    logbuffer[size] = '\0';
                    codec->stats_in = logbuffer;
                }
            }
        }
        if(codec->codec_type == CODEC_TYPE_VIDEO){
            int size= codec->width * codec->height;
            bit_buffer_size= FFMAX(bit_buffer_size, 4*size);
        }
    }

    if (!bit_buffer)
        bit_buffer = (uint8_t*)av_malloc(bit_buffer_size);
    if (!bit_buffer)
        goto fail;

    /* dump the file output parameters - cannot be done before in case
       of stream copy */
    for(i=0;i<nb_output_files;i++) {
        dump_format(output_files[i], i, output_files[i]->filename, 1);
    }

    /* dump the stream mapping */
    if (options.verbose >= 0) {
        fprintf(stderr, "Stream mapping:\n");
        for(i=0;i<nb_ostreams;i++) {
            ost = ost_table[i];
            fprintf(stderr, "  Stream #%d.%d -> #%d.%d",
                    ist_table[ost->source_index]->file_index,
                    ist_table[ost->source_index]->index,
                    ost->file_index,
                    ost->index);
            if (ost->sync_ist != ist_table[ost->source_index])
                fprintf(stderr, " [sync #%d.%d]",
                        ost->sync_ist->file_index,
                        ost->sync_ist->index);
            fprintf(stderr, "\n");
        }
    }

    /* open each encoder */
    for(i=0;i<nb_ostreams;i++) {
        ost = ost_table[i];
        if (ost->encoding_needed) {
            AVCodec *codec;
            codec = avcodec_find_encoder(ost->st->codec->codec_id);
            if (!codec) {
 //               fprintf(stderr, "Unsupported codec for output stream #%d.%d\n",
 //                       ost->file_index, ost->index);
 //               exit(1);
                throw std::exception(); // std::exception("Unsupported codec for output stream");
            }
            if (avcodec_open(ost->st->codec, codec) < 0) {
//                fprintf(stderr, "Error while opening codec for output stream #%d.%d - maybe incorrect parameters such as bit_rate, rate, width or height\n",
//                        ost->file_index, ost->index);
//                exit(1);
                throw std::exception(); // std::exception("Error while opening codec for output stream");
            }
            options.extra_size += ost->st->codec->extradata_size;
        }
    }

    /* open each decoder */
    for(i=0;i<nb_istreams;i++) {
        ist = ist_table[i];
        if (ist->decoding_needed) {
            AVCodec *codec;
            codec = avcodec_find_decoder(ist->st->codec->codec_id);
            if (!codec) {
 //               fprintf(stderr, "Unsupported codec (id=%d) for input stream #%d.%d\n",
 //                       ist->st->codec->codec_id, ist->file_index, ist->index);
 //               exit(1);
                throw std::exception(); // std::exception("Unsupported codec for input stream");
            }
            if (avcodec_open(ist->st->codec, codec) < 0) {
//                fprintf(stderr, "Error while opening codec for input stream #%d.%d\n",
//                        ist->file_index, ist->index);
//                exit(1);
                throw std::exception(); // std::exception("Error while opening codec for input stream");
            }
            //if (ist->st->codec->codec_type == CODEC_TYPE_VIDEO)
            //    ist->st->codec->flags |= CODEC_FLAG_REPEAT_FIELD;
        }
    }

    /* init pts */
    for(i=0;i<nb_istreams;i++) {
        ist = ist_table[i];
        is = input_files[ist->file_index];
        ist->pts = 0;
        ist->next_pts= AV_NOPTS_VALUE;
        ist->is_start = 1;
    }

    /* set meta data information from input file if required */
    for (i=0;i<options.nb_meta_data_maps;i++) {
        AVFormatContext *out_file;
        AVFormatContext *in_file;

        int out_file_index = options.meta_data_maps[i].out_file;
        int in_file_index = options.meta_data_maps[i].in_file;
        if ( out_file_index < 0 || out_file_index >= nb_output_files ) {
            fprintf(stderr, "Invalid output file index %d map_meta_data(%d,%d)\n", out_file_index, out_file_index, in_file_index);
            ret = AVERROR(EINVAL);
            goto fail;
        }
        if ( in_file_index < 0 || in_file_index >= nb_input_files ) {
            fprintf(stderr, "Invalid input file index %d map_meta_data(%d,%d)\n", in_file_index, out_file_index, in_file_index);
            ret = AVERROR(EINVAL);
            goto fail;
        }

        out_file = output_files[out_file_index];
        in_file = input_files[in_file_index];

        strcpy(out_file->title, in_file->title);
        strcpy(out_file->author, in_file->author);
        strcpy(out_file->copyright, in_file->copyright);
        strcpy(out_file->comment, in_file->comment);
        strcpy(out_file->album, in_file->album);
        out_file->year = in_file->year;
        out_file->track = in_file->track;
        strcpy(out_file->genre, in_file->genre);
    }

    /* open files and write file headers */
    for(i=0;i<nb_output_files;i++) {
        os = output_files[i];
        if (av_write_header(os) < 0) {
            fprintf(stderr, "Could not write header for output file #%d (incorrect codec parameters ?)\n", i);
            ret = AVERROR(EINVAL);
            goto fail;
        }
    }

    if ( !options.using_stdin && options.verbose >= 0) {
        fprintf(stderr, "Press [q] to stop encoding\n");
        url_set_interrupt_cb(decode_interrupt_cb);
    }
    term_init();

    key = -1;
    options.timer_start = av_gettime();

    for(; received_sigterm == 0;) {
        int file_index, ist_index;
        AVPacket pkt;
        double ipts_min;
        double opts_min;

    redo:
        ipts_min= 1e100;
        opts_min= 1e100;

        /* if 'q' pressed, exits */
        /*
        if (!options.using_stdin) {
            if (q_pressed)
                break;
            /!* read_key() returns 0 on EOF *!/
            key = read_key();
            if (key == 'q')
                break;
        }
        */

		// Update progress callback.
		uint64_t bytesProcessed = 0;
		for(i=0;i<nb_input_files;i++) {
			is = input_files[i];
			bytesProcessed += url_ftell(is->pb);
		}
		float percentComplete = (float)bytesProcessed / totalBytesToProcess;
		if (percentComplete - lastPercentComplete >= 0.01)
		{
			if (!progress.Update(percentComplete))
				throw std::exception(); // std::exception();
			lastPercentComplete = percentComplete;
		}

        /* select the stream that we must read now by looking at the
           smallest output pts */
        file_index = -1;
        for(i=0;i<nb_ostreams;i++) {
            double ipts, opts;
            ost = ost_table[i];
            os = output_files[ost->file_index];
            ist = ist_table[ost->source_index];
            if(ost->st->codec->codec_type == CODEC_TYPE_VIDEO)
                opts = ost->sync_opts * av_q2d(ost->st->codec->time_base);
            else
                opts = ost->st->pts.val * av_q2d(ost->st->time_base);
            ipts = (double)ist->pts;
            if (!file_table[ist->file_index].eof_reached){
                if(ipts < ipts_min) {
                    ipts_min = ipts;
                    if(options.input_sync ) file_index = ist->file_index;
                }
                if(opts < opts_min) {
                    opts_min = opts;
                    if(!options.input_sync) file_index = ist->file_index;
                }
            }
            if(ost->frame_number >= options.max_frames[ost->st->codec->codec_type]){
                file_index= -1;
                break;
            }
        }
        /* if none, if is finished */
        if (file_index < 0) {
            break;
        }

        /* finish if recording time exhausted */
        if (options.recording_time > 0 && opts_min >= (options.recording_time / 1000000.0))
            break;

        /* finish if limit size exhausted */
        if (options.limit_filesize != 0 && options.limit_filesize < static_cast<uint64_t>(url_ftell(output_files[0]->pb)))
            break;

        /* read a frame from it and output it in the fifo */
        is = input_files[file_index];
        if (av_read_frame(is, &pkt) < 0) {
            file_table[file_index].eof_reached = 1;
            if (options.opt_shortest)
                break;
            else
                continue;
        }

        if (options.do_pkt_dump) {
            av_pkt_dump_log(NULL, AV_LOG_DEBUG, &pkt, options.do_hex_dump);
        }
        /* the following test is needed in case new streams appear
           dynamically in stream : we ignore them */
        if (pkt.stream_index >= file_table[file_index].nb_streams)
            goto discard_packet;
        ist_index = file_table[file_index].ist_index + pkt.stream_index;
        ist = ist_table[ist_index];
        if (ist->discard)
            goto discard_packet;

        if (pkt.dts != AV_NOPTS_VALUE)
            pkt.dts += av_rescale_q(options.input_files_ts_offset[ist->file_index], tbq, ist->st->time_base);
        if (pkt.pts != AV_NOPTS_VALUE)
            pkt.pts += av_rescale_q(options.input_files_ts_offset[ist->file_index], tbq, ist->st->time_base);

        if(options.input_files_ts_scale[file_index][pkt.stream_index]){
            if(pkt.pts != AV_NOPTS_VALUE)
                pkt.pts *= options.input_files_ts_scale[file_index][pkt.stream_index];
            if(pkt.dts != AV_NOPTS_VALUE)
                pkt.dts *= options.input_files_ts_scale[file_index][pkt.stream_index];
        }

//        fprintf(stderr, "next:%"PRId64" dts:%"PRId64" off:%"PRId64" %d\n", ist->next_pts, pkt.dts, input_files_ts_offset[ist->file_index], ist->st->codec->codec_type);
        if (pkt.dts != AV_NOPTS_VALUE && ist->next_pts != AV_NOPTS_VALUE) {
            int64_t pkt_dts= av_rescale_q(pkt.dts, ist->st->time_base, tbq);
            int64_t delta= pkt_dts - ist->next_pts;
            if((FFABS(delta) > 1LL*options.dts_delta_threshold*AV_TIME_BASE || pkt_dts+1<ist->pts)&& !options.copy_ts){
                options.input_files_ts_offset[ist->file_index]-= delta;
//                if (options.verbose > 2)
//                    fprintf(stderr, "timestamp discontinuity %"PRId64", new offset= %"PRId64"\n", delta, options.input_files_ts_offset[ist->file_index]);
                pkt.dts-= av_rescale_q(delta, tbq, ist->st->time_base);
                if(pkt.pts != AV_NOPTS_VALUE)
                    pkt.pts-= av_rescale_q(delta, tbq, ist->st->time_base);
            }
        }

        //fprintf(stderr,"read #%d.%d size=%d\n", ist->file_index, ist->index, pkt.size);
        if (output_packet(ist, ist_index, ost_table, nb_ostreams, &pkt, options) < 0) {

            if (options.verbose >= 0)
                fprintf(stderr, "Error while decoding stream #%d.%d\n",
                        ist->file_index, ist->index);

            av_free_packet(&pkt);
            goto redo;
        }

    discard_packet:
        av_free_packet(&pkt);

        /* dump report by using the output first video and audio streams */
        print_report(output_files, ost_table, nb_ostreams, 0, options);
    }

    /* at the end of stream, we must flush the decoder buffers */
    for(i=0;i<nb_istreams;i++) {
        ist = ist_table[i];
        if (ist->decoding_needed) {
            output_packet(ist, i, ost_table, nb_ostreams, NULL, options);
        }
    }

    //term_exit();

    /* write the trailer if needed and close file */
    for(i=0;i<nb_output_files;i++) {
        os = output_files[i];
        av_write_trailer(os);
    }

    /* dump report by using the first video and audio streams */
    print_report(output_files, ost_table, nb_ostreams, 1, options);

    /* close each encoder */
    for(i=0;i<nb_ostreams;i++) {
        ost = ost_table[i];
        if (ost->encoding_needed) {
            av_freep(&ost->st->codec->stats_in);
            avcodec_close(ost->st->codec);
        }
    }

    /* close each decoder */
    for(i=0;i<nb_istreams;i++) {
        ist = ist_table[i];
        if (ist->decoding_needed) {
            avcodec_close(ist->st->codec);
        }
    }

    /* finished ! */

    ret = 0;
 fail1:
    av_freep(&bit_buffer);
    av_free(file_table);

    if (ist_table) {
        for(i=0;i<nb_istreams;i++) {
            ist = ist_table[i];
            av_free(ist);
        }
        av_free(ist_table);
    }
    if (ost_table) {
        for(i=0;i<nb_ostreams;i++) {
            ost = ost_table[i];
            if (ost) {
                if (ost->logfile) {
                    fclose(ost->logfile);
                    ost->logfile = NULL;
                }
                av_fifo_free(&ost->fifo); /* works even if fifo is not
                                             initialized but set to zero */
                av_free(ost->pict_tmp.data[0]);
                if (ost->video_resample)
                    sws_freeContext(ost->img_resample_ctx);
                if (ost->resample)
                    audio_resample_close(ost->resample);
                if (ost->reformat_ctx)
                    av_audio_convert_free(ost->reformat_ctx);
                av_free(ost);
            }
        }
        av_free(ost_table);
    }
    return ret;
 fail:
    ret = AVERROR(ENOMEM);
    goto fail1;
}

#if 0
int file_read(const char *filename)
{
    URLContext *h;
    unsigned char buffer[1024];
    int len, i;

    if (url_open(&h, filename, O_RDONLY) < 0) {
        printf("could not open '%s'\n", filename);
        return -1;
    }
    for(;;) {
        len = url_read(h, buffer, sizeof(buffer));
        if (len <= 0)
            break;
        for(i=0;i<len;i++) putchar(buffer[i]);
    }
    url_close(h);
    return 0;
}
#endif

static void opt_format(const char *arg, Options& options)
{
    /* compatibility stuff for pgmyuv */
    if (!strcmp(arg, "pgmyuv")) {
        options.pgmyuv_compatibility_hack=1;
//        opt_image_format(arg);
        arg = "image2";
        fprintf(stderr, "pgmyuv format is deprecated, use image2\n");
    }

    options.file_iformat = av_find_input_format(arg);
    options.file_oformat = guess_format(arg, NULL, NULL);
    if (!options.file_iformat && !options.file_oformat) {
 //       fprintf(stderr, "Unknown input or output format: %s\n", arg);
 //       exit(1);
        throw std::exception(); // std::exception("Unknown input or output format");
    }
}

static void opt_video_rc_override_string(char *arg, Options& options)
{
    options.video_rc_override_string = arg;
}

static void opt_me_threshold(const char *arg, Options& options)
{
    options.me_threshold = atoi(arg);
}

static void opt_verbose(const char *arg, Options& options)
{
    options.verbose = atoi(arg);
    av_log_set_level(atoi(arg));
}

static void opt_frame_rate(const char *arg, Options& options)
{
    if (av_parse_video_frame_rate(&options.frame_rate, arg) < 0) {
 //       fprintf(stderr, "Incorrect frame rate\n");
 //       exit(1);
        throw std::exception(); // std::exception("Incorrect frame rate");
    }
}

static void opt_frame_crop_top(const char *arg, Options& options)
{
    options.frame_topBand = atoi(arg);
    if (options.frame_topBand < 0) {
//        fprintf(stderr, "Incorrect top crop size\n");
//        exit(1);
        throw std::exception(); // std::exception("Incorrect top crop size");
    }
    if ((options.frame_topBand % 2) != 0) {
//        fprintf(stderr, "Top crop size must be a multiple of 2\n");
//        exit(1);
        throw std::exception(); // std::exception("Top crop size must be a multiple of 2");
    }
    if ((options.frame_topBand) >= options.frame_height){
//        fprintf(stderr, "Vertical crop dimensions are outside the range of the original image.\nRemember to crop first and scale second.\n");
//        exit(1);
        throw std::exception(); // std::exception("Vertical crop dimensions are outside the range of the original image.  Remember to crop first and scale second.");
    }
    options.frame_height -= options.frame_topBand;
}

static void opt_frame_crop_bottom(const char *arg, Options& options)
{
    options.frame_bottomBand = atoi(arg);
    if (options.frame_bottomBand < 0) {
//        fprintf(stderr, "Incorrect bottom crop size\n");
//        exit(1);
        throw std::exception(); //  std::exception("Incorrect bottom crop size");
    }
    if ((options.frame_bottomBand % 2) != 0) {
//        fprintf(stderr, "Bottom crop size must be a multiple of 2\n");
//        exit(1);
        throw std::exception(); //  std::exception("Bottom crop size must be a multiple of 2");
    }
    if ((options.frame_bottomBand) >= options.frame_height){
//        fprintf(stderr, "Vertical crop dimensions are outside the range of the original image.\nRemember to crop first and scale second.\n");
//        exit(1);
        throw std::exception(); //  std::exception("Vertical crop dimensions are outside the range of the original image.  Remember to crop first and scale second.");
    }
    options.frame_height -= options.frame_bottomBand;
}

static void opt_frame_crop_left(const char *arg, Options& options)
{
    options.frame_leftBand = atoi(arg);
    if (options.frame_leftBand < 0) {
//        fprintf(stderr, "Incorrect left crop size\n");
//        exit(1);
        throw std::exception(); // std::exception("Incorrect left crop size");
    }
    if ((options.frame_leftBand % 2) != 0) {
//        fprintf(stderr, "Left crop size must be a multiple of 2\n");
//        exit(1);
        throw std::exception(); // std::exception("Left crop size must be a multiple of 2");
    }
    if ((options.frame_leftBand) >= options.frame_width){
//        fprintf(stderr, "Horizontal crop dimensions are outside the range of the original image.\nRemember to crop first and scale second.\n");
//        exit(1);
        throw std::exception(); // std::exception("Horizontal crop dimensions are outside the range of the original image.  Remember to crop first and scale second.");
    }
    options.frame_width -= options.frame_leftBand;
}

static void opt_frame_crop_right(const char *arg, Options& options)
{
    options.frame_rightBand = atoi(arg);
    if (options.frame_rightBand < 0) {
//        fprintf(stderr, "Incorrect right crop size\n");
//        exit(1);
        throw std::exception(); // std::exception("Incorrect right crop size");
    }
    if ((options.frame_rightBand % 2) != 0) {
//        fprintf(stderr, "Right crop size must be a multiple of 2\n");
//        exit(1);
        throw std::exception(); // std::exception("Right crop size must be a multiple of 2");
    }
    if ((options.frame_rightBand) >= options.frame_width){
//        fprintf(stderr, "Horizontal crop dimensions are outside the range of the original image.\nRemember to crop first and scale second.\n");
//        exit(1);
        throw std::exception(); // std::exception("Horizontal crop dimensions are outside the range of the original image.  Remember to crop first and scale second.");
    }
    options.frame_width -= options.frame_rightBand;
}

static void opt_frame_size(const char *arg, Options& options)
{
    if (av_parse_video_frame_size(&options.frame_width, &options.frame_height, arg) < 0) {
 //       fprintf(stderr, "Incorrect frame size\n");
 //       exit(1);
        throw std::exception(); // std::exception("Incorrect frame size");
    }
    if ((options.frame_width % 2) != 0 || (options.frame_height % 2) != 0) {
        //fprintf(stderr, "Frame size must be a multiple of 2\n");
        //exit(1);
        throw std::exception(); // std::exception("Frame size must be a multiple of 2");
    }
}


#define SCALEBITS 10
#define ONE_HALF  (1 << (SCALEBITS - 1))
#define FIX(x)    ((int) ((x) * (1<<SCALEBITS) + 0.5))

#define RGB_TO_Y(r, g, b) \
((FIX(0.29900) * (r) + FIX(0.58700) * (g) + \
  FIX(0.11400) * (b) + ONE_HALF) >> SCALEBITS)

#define RGB_TO_U(r1, g1, b1, shift)\
(((- FIX(0.16874) * r1 - FIX(0.33126) * g1 +         \
     FIX(0.50000) * b1 + (ONE_HALF << shift) - 1) >> (SCALEBITS + shift)) + 128)

#define RGB_TO_V(r1, g1, b1, shift)\
(((FIX(0.50000) * r1 - FIX(0.41869) * g1 -           \
   FIX(0.08131) * b1 + (ONE_HALF << shift) - 1) >> (SCALEBITS + shift)) + 128)

static void opt_pad_color(const char *arg, Options& options) {
    /* Input is expected to be six hex digits similar to
       how colors are expressed in html tags (but without the #) */
    int rgb = strtol(arg, NULL, 16);
    int r,g,b;

    r = (rgb >> 16);
    g = ((rgb >> 8) & 255);
    b = (rgb & 255);

    options.padcolor[0] = RGB_TO_Y(r,g,b);
    options.padcolor[1] = RGB_TO_U(r,g,b,0);
    options.padcolor[2] = RGB_TO_V(r,g,b,0);
}

static void opt_frame_pad_top(const char *arg, Options& options)
{
    options.frame_padtop = atoi(arg);
    if (options.frame_padtop < 0) {
        //fprintf(stderr, "Incorrect top pad size\n");
        //exit(1);
        throw std::exception(); // std::exception("Incorrect top pad size");
    }
    if ((options.frame_padtop % 2) != 0) {
        //fprintf(stderr, "Top pad size must be a multiple of 2\n");
        //exit(1);
        throw std::exception(); // std::exception("Top pad size must be a multiple of 2");
    }
}

static void opt_frame_pad_bottom(const char *arg, Options& options)
{
    options.frame_padbottom = atoi(arg);
    if (options.frame_padbottom < 0) {
 //       fprintf(stderr, "Incorrect bottom pad size\n");
 //       exit(1);
        throw std::exception(); // std::exception("Incorrect bottom pad size");
    }
    if ((options.frame_padbottom % 2) != 0) {
//        fprintf(stderr, "Bottom pad size must be a multiple of 2\n");
//        exit(1);
        throw std::exception(); // std::exception("Bottom pad size must be a multiple of 2");
    }
}


static void opt_frame_pad_left(const char *arg, Options& options)
{
    options.frame_padleft = atoi(arg);
    if (options.frame_padleft < 0) {
//        fprintf(stderr, "Incorrect left pad size\n");
//        exit(1);
        throw std::exception(); // std::exception("Incorrect left pad size");
    }
    if ((options.frame_padleft % 2) != 0) {
//        fprintf(stderr, "Left pad size must be a multiple of 2\n");
//        exit(1);
        throw std::exception(); // std::exception("Left pad size must be a multiple of 2");
    }
}


static void opt_frame_pad_right(const char *arg, Options& options)
{
    options.frame_padright = atoi(arg);
    if (options.frame_padright < 0) {
        //fprintf(stderr, "Incorrect right pad size\n");
        //exit(1);
        throw std::exception(); // std::exception("Incorrect right pad size");
    }
    if ((options.frame_padright % 2) != 0) {
        //fprintf(stderr, "Right pad size must be a multiple of 2\n");
        //exit(1);
        throw std::exception(); // std::exception("Right pad size must be a multiple of 2");
    }
}

void list_fmts(void (*get_fmt_string)(char *buf, int buf_size, int fmt), int nb_fmts)
{
    int i;
    char fmt_str[128];
    for (i=-1; i < nb_fmts; i++) {
        get_fmt_string (fmt_str, sizeof(fmt_str), i);
        fprintf(stdout, "%s\n", fmt_str);
    }
}

static void opt_frame_pix_fmt(const char *arg, Options& options)
{
    if (strcmp(arg, "list"))
        options.frame_pix_fmt = avcodec_get_pix_fmt(arg);
    else {
        list_fmts(avcodec_pix_fmt_string, PIX_FMT_NB);
        //exit(0);
    }
}

static void opt_frame_aspect_ratio(const char *arg, Options& options)
{
    int x = 0, y = 0;
    double ar = 0;
    const char *p;

    p = strchr(arg, ':');
    if (p) {
        x = strtol(arg, (char **)&arg, 10);
        if (arg == p)
            y = strtol(arg+1, (char **)&arg, 10);
        if (x > 0 && y > 0)
            ar = (double)x / (double)y;
    } else
        ar = strtod(arg, (char **)&arg);

    if (!ar) {
        //fprintf(stderr, "Incorrect aspect ratio specification.\n");
        //exit(1);
        throw std::exception(); // std::exception("Incorrect aspect ratio specification.");
    }
    options.frame_aspect_ratio = static_cast<float>(ar);
}

static void opt_qscale(const char *arg, Options& options)
{
    options.video_qscale = static_cast<float>(atof(arg));
    if (options.video_qscale <= 0 ||
        options.video_qscale > 255) {
        //fprintf(stderr, "qscale must be > 0.0 and <= 255\n");
        //exit(1);
        throw std::exception(); // std::exception("qscale must be > 0.0 and <= 255");
    }
}

static void opt_top_field_first(const char *arg, Options& options)
{
    options.top_field_first= atoi(arg);
}

static void opt_thread_count(const char *arg, Options& options)
{
    options.thread_count= atoi(arg);
#if !defined(HAVE_THREADS)
    if (options.verbose >= 0)
        fprintf(stderr, "Warning: not compiled with thread support, using thread emulation\n");
#endif
}

static void opt_audio_sample_fmt(const char *arg, Options& options)
{
    if (strcmp(arg, "list"))
        options.audio_sample_fmt = avcodec_get_sample_fmt(arg);
    else {
        list_fmts(avcodec_sample_fmt_string, SAMPLE_FMT_NB);
        //av_exit(0);
    }
}

static void opt_audio_rate(const char *arg, Options& options)
{
    options.audio_sample_rate = atoi(arg);
}

static void opt_audio_channels(const char *arg, Options& options)
{
    options.audio_channels = atoi(arg);
}

static void opt_video_channel(const char *arg, Options& options)
{
    options.video_channel = strtol(arg, NULL, 0);
}

static void opt_video_standard(const char *arg, Options& options)
{
    options.video_standard = av_strdup(arg);
}

static void opt_codec(int *pstream_copy, char **pcodec_name,
                      int codec_type, const char *arg, Options& options)
{
    av_freep(pcodec_name);
    if (!strcmp(arg, "copy")) {
        *pstream_copy = 1;
    } else {
        *pcodec_name = av_strdup(arg);
    }
}

static void opt_audio_codec(const char *arg, Options& options)
{
    opt_codec(&options.audio_stream_copy, &options.audio_codec_name, CODEC_TYPE_AUDIO, arg, options);
}

static void opt_audio_tag(const char *arg, Options& options)
{
    char *tail;
    options.audio_codec_tag= strtol(arg, &tail, 0);

    if(!tail || *tail)
        options.audio_codec_tag= arg[0] + (arg[1]<<8) + (arg[2]<<16) + (arg[3]<<24);
}

static void opt_video_tag(const char *arg, Options& options)
{
    char *tail;
    options.video_codec_tag= strtol(arg, &tail, 0);

    if(!tail || *tail)
        options.video_codec_tag= arg[0] + (arg[1]<<8) + (arg[2]<<16) + (arg[3]<<24);
}

#ifdef CONFIG_VHOOK
static void add_frame_hooker(const char *arg)
{
    int argc = 0;
    char *argv[64];
    int i;
    char *args = av_strdup(arg);

    using_vhook = 1;

    argv[0] = strtok(args, " ");
    while (argc < 62 && (argv[++argc] = strtok(NULL, " "))) {
    }

    i = frame_hook_add(argc, argv);

    if (i != 0) {
        fprintf(stderr, "Failed to add video hook function: %s\n", arg);
        exit(1);
    }
}
#endif

static void opt_video_codec(const char *arg, Options& options)
{
    opt_codec(&options.video_stream_copy, &options.video_codec_name, CODEC_TYPE_VIDEO, arg, options);
}

static void opt_subtitle_codec(const char *arg, Options& options)
{
    opt_codec(&options.subtitle_stream_copy, &options.subtitle_codec_name, CODEC_TYPE_SUBTITLE, arg, options);
}

static void opt_map(const char *arg, Options& options)
{
    AVStreamMap *m;
    const char *p;

    p = arg;
    m = &options.stream_maps[options.nb_stream_maps++];

    m->file_index = strtol(arg, (char **)&p, 0);
    if (*p)
        p++;

    m->stream_index = strtol(p, (char **)&p, 0);
    if (*p) {
        p++;
        m->sync_file_index = strtol(p, (char **)&p, 0);
        if (*p)
            p++;
        m->sync_stream_index = strtol(p, (char **)&p, 0);
    } else {
        m->sync_file_index = m->file_index;
        m->sync_stream_index = m->stream_index;
    }
}

static void opt_map_meta_data(const char *arg, Options& options)
{
    AVMetaDataMap *m;
    const char *p;

    p = arg;
    m = &options.meta_data_maps[options.nb_meta_data_maps++];

    m->out_file = strtol(arg, (char **)&p, 0);
    if (*p)
        p++;

    m->in_file = strtol(p, (char **)&p, 0);
}

static void opt_input_ts_scale(const char *arg, Options& options)
{
    unsigned int stream;
    double scale;
    char *p;

    stream = strtol(arg, &p, 0);
    if (*p)
        p++;
    scale= strtod(p, &p);

    if(stream >= MAX_STREAMS)
        throw std::exception();

    options.input_files_ts_scale[options.nb_input_files][stream]= scale;
}

static void opt_recording_time(const char *arg, Options& options)
{
    options.recording_time = parse_date(arg, 1);
}

static void opt_limit_filesize(const char *arg, Options& options)
{
    options.limit_filesize = _strtoi64(arg, (char **)NULL, 10);
}

static void opt_start_time(const char *arg, Options& options)
{
    options.start_time = parse_date(arg, 1);
}

static void opt_rec_timestamp(const char *arg, Options& options)
{
    options.rec_timestamp = parse_date(arg, 0) / 1000000;
}

static void opt_input_ts_offset(const char *arg, Options& options)
{
    options.input_ts_offset = parse_date(arg, 1);
}

static void opt_str_title(const char *arg, Options& options)
{
    options.str_title = av_strdup(arg);
}

static void opt_str_author(const char *arg, Options& options)
{
    options.str_author = av_strdup(arg);
}

static void opt_str_copyright(const char *arg, Options& options)
{
    options.str_copyright = av_strdup(arg);
}

static void opt_str_comment(const char *arg, Options& options)
{
    options.str_comment = av_strdup(arg);
}

static void opt_str_genre(const char *arg, Options& options)
{
    options.str_genre = av_strdup(arg);
}

static void opt_str_album(const char *arg, Options& options)
{
    options.str_album = av_strdup(arg);
}

static void opt_do_benchmark(const char *arg, Options& options)
{
    options.do_benchmark = 1;
}

static void opt_do_pkt_dump(const char *arg, Options& options)
{
    options.do_pkt_dump = 1;
}

static void opt_do_hex_dump(const char *arg, Options& options)
{
    options.do_hex_dump = 1;
}

static void opt_rate_emu(const char *arg, Options& options)
{
    options.rate_emu = 1;
}

static void opt_loop_input(const char *arg, Options& options)
{
    options.loop_input = 1;
}

static void opt_loop_output(const char *arg, Options& options)
{
    options.loop_output = atoi(arg);
}

static void opt_video_sync_method(const char *arg, Options& options)
{
    options.video_sync_method = atoi(arg);
}

static void opt_audio_sync_method(const char *arg, Options& options)
{
    options.audio_sync_method = atoi(arg);
}

static void opt_audio_drift_threshold(const char *arg, Options& options)
{
    options.audio_drift_threshold = static_cast<float>(atof(arg));
}

static void opt_video_global_header(const char *arg, Options& options)
{
    options.video_global_header = atoi(arg);
}

static void opt_copy_ts(const char *arg, Options& options)
{
    options.copy_ts = 1;
}

static void opt_shortest(const char *arg, Options& options)
{
    options.opt_shortest = 1;
}

static void opt_dts_delta_threshold(const char *arg, Options& options)
{
    options.dts_delta_threshold = static_cast<float>(atof(arg));
}

static void opt_programid(const char *arg, Options& options)
{
    options.opt_programid = atoi(arg);
}

static void opt_max_frames_video(const char *arg, Options& options)
{
    options.max_frames[CODEC_TYPE_VIDEO] = atoi(arg);
}

static void opt_max_frames_data(const char *arg, Options& options)
{
    options.max_frames[CODEC_TYPE_DATA] = atoi(arg);
}

static void opt_intra_only(const char *arg, Options& options)
{
    options.intra_only = 1;
}

static void opt_video_disable(const char *arg, Options& options)
{
    options.video_disable = 1;
}

static void opt_video_discard(const char *arg, Options& options)
{
    options.video_discard = atoi(arg);
}

static void opt_same_quality(const char *arg, Options& options)
{
    options.same_quality = 1;
}

static void opt_pass_logfilename(const char *arg, Options& options)
{
    options.pass_logfilename = av_strdup(arg);
}

static void opt_do_deinterlace(const char *arg, Options& options)
{
    options.do_deinterlace = 1;
}

static void opt_do_psnr(const char *arg, Options& options)
{
    options.do_psnr = 1;
}

static void opt_intra_dc_precision(const char *arg, Options& options)
{
    options.intra_dc_precision = atoi(arg);
}

static void opt_qp_hist(const char *arg, Options& options)
{
    options.qp_hist = 1;
}

static void opt_max_frames_audio(const char *arg, Options& options)
{
    options.max_frames[CODEC_TYPE_AUDIO] = atoi(arg);
}

static void opt_audio_qscale(const char *arg, Options& options)
{
    options.audio_qscale = static_cast<float>(atof(arg));
}

static void opt_audio_disable(const char *arg, Options& options)
{
    options.audio_disable = 1;
}

static void opt_audio_volume(const char *arg, Options& options)
{
    options.audio_volume = atoi(arg);
}

static void opt_audio_language(const char *arg, Options& options)
{
    options.audio_language = av_strdup(arg);
}

static void opt_subtitle_disable(const char *arg, Options& options)
{
    options.subtitle_disable = 1;
}

static void opt_subtitle_language(const char *arg, Options& options)
{
    options.subtitle_language = av_strdup(arg);
}

static void opt_input_sync(const char *arg, Options& options)
{
    options.input_sync = 1;
}

static void opt_mux_max_delay(const char *arg, Options& options)
{
    options.mux_max_delay = static_cast<float>(atof(arg));
}

static void opt_mux_preload(const char *arg, Options& options)
{
    options.mux_preload = static_cast<float>(atof(arg));
}

static enum CodecID find_codec_or_die(const char *name, int type, int encoder)
{
    AVCodec *codec;

    if(!name)
        return CODEC_ID_NONE;
    codec = encoder ?
        avcodec_find_encoder_by_name(name) :
        avcodec_find_decoder_by_name(name);
    if(!codec) {
        //av_log(NULL, AV_LOG_ERROR, "Unknown codec '%s'\n", name);
        //exit(1);
        throw std::exception(); // std::exception("Unknown codec");
    }
    if(codec->type != type) {
        //av_log(NULL, AV_LOG_ERROR, "Invalid codec type '%s'\n", name);
        //exit(1);
        throw std::exception(); // std::exception("Invalid codec type");
    }
    return codec->id;
}

static void set_context_opts(void *ctx, void *opts_ctx, int flags, Options& options)
{
    int i;
    for(i=0; i<options.opt_name_count; i++){
        char buf[256];
        const AVOption *opt;
        const char *str= av_get_string(opts_ctx, options.opt_names[i], &opt, buf, sizeof(buf));
        /* if an option with name opt_names[i] is present in opts_ctx then str is non-NULL */
        if(str && ((opt->flags & flags) == flags))
            av_set_string3(ctx, options.opt_names[i], str, 1, NULL);
    }
}

static void opt_input_file(const char *filename, Options& options)
{
    AVFormatContext *ic;
    AVFormatParameters params, *ap = &params;
    int err, i, ret, rfps, rfps_base;
    int64_t timestamp;

    if (!strcmp(filename, "-"))
        filename = "pipe:";

    options.using_stdin |= !strncmp(filename, "pipe:", 5) ||
                   !strcmp( filename, "/dev/stdin" );

    /* get default parameters from command line */
    ic = avformat_alloc_context();

    memset(ap, 0, sizeof(*ap));
    ap->prealloced_context = 1;
    ap->sample_rate = options.audio_sample_rate;
    ap->channels = options.audio_channels;
    ap->time_base.den = options.frame_rate.num;
    ap->time_base.num = options.frame_rate.den;
    ap->width = options.frame_width + options.frame_padleft + options.frame_padright;
    ap->height = options.frame_height + options.frame_padtop + options.frame_padbottom;
    ap->pix_fmt = options.frame_pix_fmt;
    // ap->sample_fmt = audio_sample_fmt; //FIXME:not implemented in libavformat
    ap->channel = options.video_channel;
    ap->standard = options.video_standard;
    ap->video_codec_id = find_codec_or_die(options.video_codec_name, CODEC_TYPE_VIDEO, 0);
    ap->audio_codec_id = find_codec_or_die(options.audio_codec_name, CODEC_TYPE_AUDIO, 0);
    if(options.pgmyuv_compatibility_hack)
        ap->video_codec_id= CODEC_ID_PGMYUV;
    
    set_context_opts(ic, options.avformat_opts, AV_OPT_FLAG_DECODING_PARAM, options);
    
    ic->video_codec_id   = find_codec_or_die(options.video_codec_name   , CODEC_TYPE_VIDEO   , 0);
    ic->audio_codec_id   = find_codec_or_die(options.audio_codec_name   , CODEC_TYPE_AUDIO   , 0);
    ic->subtitle_codec_id= find_codec_or_die(options.subtitle_codec_name, CODEC_TYPE_SUBTITLE, 0);
    
    /* open the input file with generic libav function */
    err = av_open_input_file(&ic, filename, options.file_iformat, 0, ap);
    if (err < 0) {
        throw std::exception(); // std::exception(get_error_string(filename, err).c_str());
    }
    if(options.opt_programid) {
        int i;
        for(i=0; i<ic->nb_programs; i++)
            if(ic->programs[i]->id != options.opt_programid)
                ic->programs[i]->discard = AVDISCARD_ALL;
    }

	ic->loop_input = options.loop_input;

    /* If not enough info to get the stream parameters, we decode the
       first frames to get it. (used in mpeg case for example) */
    ret = av_find_stream_info(ic);
    if (ret < 0 && options.verbose >= 0) {
        //fprintf(stderr, "%s: could not find codec parameters\n", filename);
        //exit(1);
        av_close_input_file(ic);
        std::string errMsg = std::string(filename) + "%s: could not find codec parameters";
        throw std::exception(); // std::exception(errMsg.c_str());
    }

    timestamp = options.start_time;
    /* add the stream start time */
    if (ic->start_time != AV_NOPTS_VALUE)
        timestamp += ic->start_time;

    /* if seeking requested, we execute it */
    if (options.start_time != 0) {
        ret = av_seek_frame(ic, -1, timestamp, AVSEEK_FLAG_BACKWARD);
        if (ret < 0) {
            fprintf(stderr, "%s: could not seek to position %0.3f\n",
                    filename, (double)timestamp / AV_TIME_BASE);
        }
        /* reset seek info */
        options.start_time = 0;
    }

    /* update the current parameters so that they match the one of the input stream */
    for(i=0;i<static_cast<int>(ic->nb_streams);i++) {
        AVCodecContext *enc = ic->streams[i]->codec;
        if(options.thread_count>1)
            avcodec_thread_init(enc, options.thread_count);
        enc->thread_count= options.thread_count;
        switch(enc->codec_type) {
        case CODEC_TYPE_AUDIO:
            set_context_opts(enc, options.avctx_opts[CODEC_TYPE_AUDIO], AV_OPT_FLAG_AUDIO_PARAM | AV_OPT_FLAG_DECODING_PARAM, options);
            //fprintf(stderr, "\nInput Audio channels: %d", enc->channels);
            options.audio_channels = enc->channels;
            options.audio_sample_rate = enc->sample_rate;
            options.audio_sample_fmt = enc->sample_fmt;
            if(options.audio_disable)
                ic->streams[i]->discard= AVDISCARD_ALL;
            break;
        case CODEC_TYPE_VIDEO:
            set_context_opts(enc, options.avctx_opts[CODEC_TYPE_VIDEO], AV_OPT_FLAG_VIDEO_PARAM | AV_OPT_FLAG_DECODING_PARAM, options);
            options.frame_height = enc->height;
            options.frame_width = enc->width;
            if(ic->streams[i]->sample_aspect_ratio.num)
                options.frame_aspect_ratio=av_q2d(ic->streams[i]->sample_aspect_ratio);
            else
                options.frame_aspect_ratio=av_q2d(enc->sample_aspect_ratio);
            options.frame_aspect_ratio *= (float) enc->width / enc->height;
            //options.frame_aspect_ratio = static_cast<float>(av_q2d(enc->sample_aspect_ratio) * enc->width / enc->height);
            options.frame_pix_fmt = enc->pix_fmt;
            rfps      = ic->streams[i]->r_frame_rate.num;
            rfps_base = ic->streams[i]->r_frame_rate.den;
            if(enc->lowres) enc->flags |= CODEC_FLAG_EMU_EDGE;
            if(options.me_threshold)
                enc->debug |= FF_DEBUG_MV;

            if (enc->time_base.den != rfps || enc->time_base.num != rfps_base) {

                if (options.verbose >= 0)
                    fprintf(stderr,"\nSeems stream %d codec frame rate differs from container frame rate: %2.2f (%d/%d) -> %2.2f (%d/%d)\n",
                            i, (float)enc->time_base.den / enc->time_base.num, enc->time_base.den, enc->time_base.num,

                    (float)rfps / rfps_base, rfps, rfps_base);
            }
            /* update the current frame rate to match the stream frame rate */
            options.frame_rate.num = rfps;
            options.frame_rate.den = rfps_base;

            enc->rate_emu = options.rate_emu;
            if(options.video_disable)
                ic->streams[i]->discard= AVDISCARD_ALL;
            else if(options.video_discard)
                ic->streams[i]->discard= (AVDiscard)options.video_discard;
            break;
        case CODEC_TYPE_DATA:
            break;
        case CODEC_TYPE_SUBTITLE:
            if(options.subtitle_disable)
                ic->streams[i]->discard = AVDISCARD_ALL;
            break;
        case CODEC_TYPE_UNKNOWN:
            break;
        default:
			ThrowInternalError();
        }
    }

    options.input_files[options.nb_input_files] = ic;
    options.input_files_ts_offset[options.nb_input_files] = options.input_ts_offset - (options.copy_ts ? 0 : timestamp);
    for( i = 0; i < MAX_STREAMS; i++ )
        options.input_files_ts_scale[options.nb_input_files][i] = 0;
    /* dump the file content */
    if (options.verbose >= 0)
        dump_format(ic, options.nb_input_files, filename, 0);

    options.nb_input_files++;
    options.file_iformat = NULL;
    options.file_oformat = NULL;

    options.video_channel = 0;

    options.rate_emu = 0;
}

static void opt_file_overwrite(const char* arg, Options& options)
{
	options.file_overwrite = true;
}

static void check_audio_video_sub_inputs(int *has_video_ptr, int *has_audio_ptr,
                                         int *has_subtitle_ptr, Options& options)
{
    int has_video, has_audio, has_subtitle, i, j;
    AVFormatContext *ic;

    has_video = 0;
    has_audio = 0;
    has_subtitle = 0;
    for(j=0;j<options.nb_input_files;j++) {
        ic = options.input_files[j];
        for(i=0;i<static_cast<int>(ic->nb_streams);i++) {
            AVCodecContext *enc = ic->streams[i]->codec;
            switch(enc->codec_type) {
            case CODEC_TYPE_AUDIO:
                has_audio = 1;
                break;
            case CODEC_TYPE_VIDEO:
                has_video = 1;
                break;
            case CODEC_TYPE_SUBTITLE:
                has_subtitle = 1;
                break;
            case CODEC_TYPE_DATA:
            case CODEC_TYPE_UNKNOWN:
                break;
            default:
                ThrowInternalError();
            }
        }
    }
    *has_video_ptr = has_video;
    *has_audio_ptr = has_audio;
    *has_subtitle_ptr = has_subtitle;
}

static void new_video_stream(AVFormatContext *oc, Options& options)
{
    AVStream *st;
    AVCodecContext *video_enc;
    int codec_id;

    st = av_new_stream(oc, oc->nb_streams);
    if (!st) {
        //fprintf(stderr, "Could not alloc stream\n");
        //exit(1);
        throw std::exception(); // std::exception("Could not alloc stream");
    }
    avcodec_get_context_defaults2(st->codec, CODEC_TYPE_VIDEO);
    options.bitstream_filters[options.nb_output_files][oc->nb_streams - 1]= options.video_bitstream_filters;
    options.video_bitstream_filters= NULL;

    if(options.thread_count>1)
        avcodec_thread_init(st->codec, options.thread_count);

    video_enc = st->codec;

    if(options.video_codec_tag)
        video_enc->codec_tag= options.video_codec_tag;

    if(   (options.video_global_header&1)
       || (options.video_global_header==0 && (oc->oformat->flags & AVFMT_GLOBALHEADER))){
        video_enc->flags |= CODEC_FLAG_GLOBAL_HEADER;
        options.avctx_opts[CODEC_TYPE_VIDEO]->flags|= CODEC_FLAG_GLOBAL_HEADER;
    }
    if(options.video_global_header&2){
        video_enc->flags2 |= CODEC_FLAG2_LOCAL_HEADER;
        options.avctx_opts[CODEC_TYPE_VIDEO]->flags2|= CODEC_FLAG2_LOCAL_HEADER;
    }

    if (options.video_stream_copy) {
        st->stream_copy = 1;
        video_enc->codec_type = CODEC_TYPE_VIDEO;
        video_enc->sample_aspect_ratio =
        st->sample_aspect_ratio = av_d2q(options.frame_aspect_ratio*options.frame_height/options.frame_width, 255);
    } else {
        char *p;
        int i;
        AVCodec *codec;

        codec_id = av_guess_codec(oc->oformat, NULL, oc->filename, NULL, CODEC_TYPE_VIDEO);
        if (options.video_codec_name)
            codec_id = find_codec_or_die(options.video_codec_name, CODEC_TYPE_VIDEO, 1);

        video_enc->codec_id = (CodecID)codec_id;
        codec = avcodec_find_encoder((CodecID)codec_id);

        set_context_opts(video_enc, options.avctx_opts[CODEC_TYPE_VIDEO], AV_OPT_FLAG_VIDEO_PARAM | AV_OPT_FLAG_ENCODING_PARAM, options);

        video_enc->time_base.den = options.frame_rate.num;
        video_enc->time_base.num = options.frame_rate.den;
        if(codec && codec->supported_framerates){
            const AVRational *p= codec->supported_framerates;
            AVRational req= {options.frame_rate.num, options.frame_rate.den};
            const AVRational *best=NULL;
            AVRational best_error= {INT_MAX, 1};
            for(; p->den!=0; p++){
                AVRational error= av_sub_q(req, *p);
                if(error.num <0) error.num *= -1;
                if(av_cmp_q(error, best_error) < 0){
                    best_error= error;
                    best= p;
                }
            }
            video_enc->time_base.den= best->num;
            video_enc->time_base.num= best->den;
        }

        video_enc->width = options.frame_width + options.frame_padright + options.frame_padleft;
        video_enc->height = options.frame_height + options.frame_padtop + options.frame_padbottom;
        video_enc->sample_aspect_ratio = av_d2q(options.frame_aspect_ratio*video_enc->height/video_enc->width, 255);
        video_enc->pix_fmt = options.frame_pix_fmt;
        st->sample_aspect_ratio = video_enc->sample_aspect_ratio;

        if(codec && codec->pix_fmts){
            const enum PixelFormat *p= codec->pix_fmts;
            for(; *p!=-1; p++){
                if(*p == video_enc->pix_fmt)
                    break;
            }
            if(*p == -1)
                video_enc->pix_fmt = codec->pix_fmts[0];
        }

        if (options.intra_only)
            video_enc->gop_size = 0;
        if (options.video_qscale || options.same_quality) {
            video_enc->flags |= CODEC_FLAG_QSCALE;
            st->quality = FF_QP2LAMBDA * options.video_qscale;
            video_enc->global_quality= static_cast<int>(st->quality);
        }

        if(options.intra_matrix)
            video_enc->intra_matrix = options.intra_matrix;
        if(options.inter_matrix)
            video_enc->inter_matrix = options.inter_matrix;

        video_enc->thread_count = options.thread_count;
        p= options.video_rc_override_string;
        for(i=0; p; i++){
            int start, end, q;
            int e=sscanf(p, "%d,%d,%d", &start, &end, &q);
            if(e!=3){
                //fprintf(stderr, "error parsing rc_override\n");
                //exit(1);
                throw std::exception(); // std::exception("error parsing rc_override");
            }
            video_enc->rc_override=
                (RcOverride*)av_realloc(video_enc->rc_override,
                           sizeof(RcOverride)*(i+1));
            video_enc->rc_override[i].start_frame= start;
            video_enc->rc_override[i].end_frame  = end;
            if(q>0){
                video_enc->rc_override[i].qscale= q;
                video_enc->rc_override[i].quality_factor= 1.0;
            }
            else{
                video_enc->rc_override[i].qscale= 0;
                video_enc->rc_override[i].quality_factor= static_cast<float>(-q/100.0);
            }
            p= strchr(p, '/');
            if(p) p++;
        }
        video_enc->rc_override_count=i;
        if (!video_enc->rc_initial_buffer_occupancy)
            video_enc->rc_initial_buffer_occupancy = video_enc->rc_buffer_size*3/4;
        video_enc->me_threshold= options.me_threshold;
        video_enc->intra_dc_precision= options.intra_dc_precision - 8;

        if (options.do_psnr)
            video_enc->flags|= CODEC_FLAG_PSNR;

        /* two pass mode */
        if (options.do_pass) {
            if (options.do_pass == 1) {
                video_enc->flags |= CODEC_FLAG_PASS1;
            } else {
                video_enc->flags |= CODEC_FLAG_PASS2;
            }
        }
    }

    /* reset some key parameters */
    options.video_disable = 0;
    av_freep(&options.video_codec_name);
    options.video_stream_copy = 0;
}

static void new_audio_stream(AVFormatContext *oc, Options& options)
{
    AVStream *st;
    AVCodecContext *audio_enc;
    int codec_id;

    st = av_new_stream(oc, oc->nb_streams);
    if (!st) {
        //fprintf(stderr, "Could not alloc stream\n");
        //exit(1);
        throw std::exception(); // std::exception("Could not alloc stream");
    }
    avcodec_get_context_defaults2(st->codec, CODEC_TYPE_AUDIO);

    options.bitstream_filters[options.nb_output_files][oc->nb_streams - 1]= options.audio_bitstream_filters;
    options.audio_bitstream_filters= NULL;

    if(options.thread_count>1)
        avcodec_thread_init(st->codec, options.thread_count);

    audio_enc = st->codec;
    audio_enc->codec_type = CODEC_TYPE_AUDIO;

    if(options.audio_codec_tag)
        audio_enc->codec_tag= options.audio_codec_tag;

    if (oc->oformat->flags & AVFMT_GLOBALHEADER) {
        audio_enc->flags |= CODEC_FLAG_GLOBAL_HEADER;
        options.avctx_opts[CODEC_TYPE_AUDIO]->flags|= CODEC_FLAG_GLOBAL_HEADER;
    }
    if (options.audio_stream_copy) {
        st->stream_copy = 1;
        audio_enc->channels = options.audio_channels;
    } else {
        AVCodec *codec;
        codec_id = av_guess_codec(oc->oformat, NULL, oc->filename, NULL, CODEC_TYPE_AUDIO);

        set_context_opts(audio_enc, options.avctx_opts[CODEC_TYPE_AUDIO], AV_OPT_FLAG_AUDIO_PARAM | AV_OPT_FLAG_ENCODING_PARAM, options);

        if (options.audio_codec_name)
            codec_id = find_codec_or_die(options.audio_codec_name, CODEC_TYPE_AUDIO, 1);
        audio_enc->codec_id = (CodecID)codec_id;
        codec = avcodec_find_encoder((CodecID)codec_id);

        if (options.audio_qscale > QSCALE_NONE) {
            audio_enc->flags |= CODEC_FLAG_QSCALE;
            st->quality = FF_QP2LAMBDA * options.audio_qscale;
            audio_enc->global_quality = static_cast<int>(st->quality);
        }
        audio_enc->thread_count = options.thread_count;
        audio_enc->channels = options.audio_channels;
        audio_enc->sample_fmt = options.audio_sample_fmt;

        if(codec && codec->sample_fmts){
            const enum SampleFormat *p= codec->sample_fmts;
            for(; *p!=-1; p++){
                if(*p == audio_enc->sample_fmt)
                    break;
            }
            if(*p == -1)
                audio_enc->sample_fmt = codec->sample_fmts[0];
        }
    }
    audio_enc->sample_rate = options.audio_sample_rate;
    AVRational tmpRational = {1, options.audio_sample_rate};
    audio_enc->time_base= tmpRational;
    if (options.audio_language) {
        av_strlcpy(st->language, options.audio_language, sizeof(st->language));
        av_free(options.audio_language);
        options.audio_language = NULL;
    }

    /* reset some key parameters */
    options.audio_disable = 0;
    av_freep(&options.audio_codec_name);
    options.audio_stream_copy = 0;
}

static void new_subtitle_stream(AVFormatContext *oc, Options& options)
{
    AVStream *st;
    AVCodecContext *subtitle_enc;

    st = av_new_stream(oc, oc->nb_streams);
    if (!st) {
        //fprintf(stderr, "Could not alloc stream\n");
        //exit(1);
        throw std::exception(); // std::exception("Could not alloc stream");
    }
    avcodec_get_context_defaults2(st->codec, CODEC_TYPE_SUBTITLE);

    subtitle_enc = st->codec;
    subtitle_enc->codec_type = CODEC_TYPE_SUBTITLE;
    if (options.subtitle_stream_copy) {
        st->stream_copy = 1;
    } else {
        set_context_opts(options.avctx_opts[CODEC_TYPE_SUBTITLE], subtitle_enc, AV_OPT_FLAG_SUBTITLE_PARAM | AV_OPT_FLAG_ENCODING_PARAM, options);
        subtitle_enc->codec_id = find_codec_or_die(options.subtitle_codec_name, CODEC_TYPE_SUBTITLE, 1);
    }

    if (options.subtitle_language) {
        av_strlcpy(st->language, options.subtitle_language, sizeof(st->language));
        av_free(options.subtitle_language);
        options.subtitle_language = NULL;
    }

    options.subtitle_disable = 0;
    av_freep(&options.subtitle_codec_name);
    options.subtitle_stream_copy = 0;
}

static void opt_new_audio_stream(const char* arg, Options& options)
{
    AVFormatContext *oc;
    if (options.nb_output_files <= 0) {
        //fprintf(stderr, "At least one output file must be specified\n");
        //exit(1);
        throw std::exception(); // std::exception("At least one output file must be specified");
    }
    oc = options.output_files[options.nb_output_files - 1];
    new_audio_stream(oc, options);
}

static void opt_new_video_stream(const char* arg, Options& options)
{
    AVFormatContext *oc;
    if (options.nb_output_files <= 0) {
        //fprintf(stderr, "At least one output file must be specified\n");
        //exit(1);
        throw std::exception(); // std::exception("At least one output file must be specified");
    }
    oc = options.output_files[options.nb_output_files - 1];
    new_video_stream(oc, options);
}

static void opt_new_subtitle_stream(const char* arg, Options& options)
{
    AVFormatContext *oc;
    if (options.nb_output_files <= 0) {
        //fprintf(stderr, "At least one output file must be specified\n");
        //exit(1);
        throw std::exception(); // std::exception("At least one output file must be specified");
    }
    oc = options.output_files[options.nb_output_files - 1];
    new_subtitle_stream(oc, options);
}

static void opt_output_file(const char *filename, Options& options)
{
    AVFormatContext *oc;
    int use_video, use_audio, use_subtitle;
    int input_has_video, input_has_audio, input_has_subtitle;
    AVFormatParameters params, *ap = &params;

    if (!strcmp(filename, "-"))
        filename = "pipe:";

    oc = avformat_alloc_context();

    if (!options.file_oformat) {
        options.file_oformat = guess_format(NULL, filename, NULL);
        if (!options.file_oformat) {
            //fprintf(stderr, "Unable to find a suitable output format for '%s'\n",
            //        filename);
            //exit(1);
            std::string errMsg("Unable to find a suitable output format for ");
            errMsg += filename;
            throw std::exception(); // std::exception(errMsg.c_str());
        }
    }

    oc->oformat = options.file_oformat;
    av_strlcpy(oc->filename, filename, sizeof(oc->filename));

    if (!strcmp(options.file_oformat->name, "ffm") &&
        av_strstart(filename, "http:", NULL)) {
        /* special case for files sent to ffserver: we get the stream
           parameters from ffserver */
        if (read_ffserver_streams(oc, filename, options) < 0) {
            //fprintf(stderr, "Could not read stream parameters from '%s'\n", filename);
            //exit(1);
            std::string errMsg("Could not read stream parameters from ");
            errMsg += filename;
            throw std::exception(); // std::exception(errMsg.c_str());
        }
    } else {
        use_video = options.file_oformat->video_codec != CODEC_ID_NONE || options.video_stream_copy || options.video_codec_name;
        use_audio = options.file_oformat->audio_codec != CODEC_ID_NONE || options.audio_stream_copy || options.audio_codec_name;
        use_subtitle = options.file_oformat->subtitle_codec != CODEC_ID_NONE || options.subtitle_stream_copy || options.subtitle_codec_name;

        /* disable if no corresponding type found and at least one
           input file */
        if (options.nb_input_files > 0) {
            check_audio_video_sub_inputs(&input_has_video, &input_has_audio,
                                         &input_has_subtitle, options);
            if (!input_has_video)
                use_video = 0;
            if (!input_has_audio)
                use_audio = 0;
            if (!input_has_subtitle)
                use_subtitle = 0;
        }

        /* manual disable */
        if (options.audio_disable) {
            use_audio = 0;
        }
        if (options.video_disable) {
            use_video = 0;
        }
        if (options.subtitle_disable) {
            use_subtitle = 0;
        }

        if (use_video) {
            new_video_stream(oc, options);
        }

        if (use_audio) {
            new_audio_stream(oc, options);
        }

        if (use_subtitle) {
            new_subtitle_stream(oc, options);
        }

        oc->timestamp = options.rec_timestamp;

        if (options.str_title)
            av_strlcpy(oc->title, options.str_title, sizeof(oc->title));
        if (options.str_author)
            av_strlcpy(oc->author, options.str_author, sizeof(oc->author));
        if (options.str_copyright)
            av_strlcpy(oc->copyright, options.str_copyright, sizeof(oc->copyright));
        if (options.str_comment)
            av_strlcpy(oc->comment, options.str_comment, sizeof(oc->comment));
        if (options.str_album)
            av_strlcpy(oc->album, options.str_album, sizeof(oc->album));
        if (options.str_genre)
            av_strlcpy(oc->genre, options.str_genre, sizeof(oc->genre));
    }

    options.output_files[options.nb_output_files++] = oc;

    /* check filename in case of an image number is expected */
    if (oc->oformat->flags & AVFMT_NEEDNUMBER) {
        if (!av_filename_number_test(oc->filename)) {
            //print_error(oc->filename, AVERROR_NUMEXPECTED);
            //exit(1);
            throw std::exception(); // std::exception(get_error_string(oc->filename, AVERROR_NUMEXPECTED).c_str());
        }
    }

    if (!(oc->oformat->flags & AVFMT_NOFILE)) {
/*
        /!* test if it already exists to avoid loosing precious files *!/
        if (!file_overwrite &&
            (strchr(filename, ':') == NULL ||
             av_strstart(filename, "file:", NULL))) {
            if (url_exist(filename)) {
                int c;

                if ( !using_stdin ) {
                    fprintf(stderr,"File '%s' already exists. Overwrite ? [y/N] ", filename);
                    fflush(stderr);
                    c = getchar();
                    if (toupper(c) != 'Y') {
                        fprintf(stderr, "Not overwriting - exiting\n");
                        exit(1);
                    }
                                }
                                else {
                    fprintf(stderr,"File '%s' already exists. Exiting.\n", filename);
                    exit(1);
                                }
            }
        }
*/

        /* open the file */
        if (url_fopen(&oc->pb, filename, URL_WRONLY) < 0) {
            //fprintf(stderr, "Could not open '%s'\n", filename);
            //exit(1);
            std::string errMsg("Could not open ");
            errMsg += filename;
            throw std::exception(); // std::exception(errMsg.c_str());
        }
    }

    memset(ap, 0, sizeof(*ap));
    if (av_set_parameters(oc, ap) < 0) {
        //fprintf(stderr, "%s: Invalid encoding parameters\n",
        //        oc->filename);
        //exit(1);
        std::string errMsg(oc->filename);
        errMsg += ": Invalid encoding parameters";
        throw std::exception(); // std::exception(errMsg.c_str());
    }

    oc->preload= (int)(options.mux_preload*AV_TIME_BASE);
    oc->max_delay= (int)(options.mux_max_delay*AV_TIME_BASE);
    oc->loop_output = options.loop_output;

    set_context_opts(oc, options.avformat_opts, AV_OPT_FLAG_ENCODING_PARAM, options);

    /* reset some options */
    options.file_oformat = NULL;
    options.file_iformat = NULL;
}

/* same option as mencoder */
static void opt_pass(const char *pass_str, Options& options)
{
    int pass;
    pass = atoi(pass_str);
    if (pass != 1 && pass != 2) {
        //fprintf(stderr, "pass number can be only 1 or 2\n");
        //exit(1);
        throw std::exception(); // std::exception("pass number can be only 1 or 2");
    }
    options.do_pass = pass;
}

static int64_t getutime(void)
{
#ifdef HAVE_GETRUSAGE
    struct rusage rusage;

    getrusage(RUSAGE_SELF, &rusage);
    return (rusage.ru_utime.tv_sec * 1000000LL) + rusage.ru_utime.tv_usec;
#elif defined(HAVE_GETPROCESSTIMES)
    HANDLE proc;
    FILETIME c, e, k, u;
    proc = GetCurrentProcess();
    GetProcessTimes(proc, &c, &e, &k, &u);
    return ((int64_t) u.dwHighDateTime << 32 | u.dwLowDateTime) / 10;
#else
    return av_gettime();
#endif
}

static void parse_matrix_coeffs(uint16_t *dest, const char *str)
{
    int i;
    const char *p = str;
    for(i = 0;; i++) {
        dest[i] = atoi(p);
        if(i == 63)
            break;
        p = strchr(p, ',');
        if(!p) {
            //fprintf(stderr, "Syntax error in matrix \"%s\" at coeff %d\n", str, i);
            //exit(1);
			std::stringstream errMsg;
			errMsg << "Syntax error in matrix \"" << str << "\" at coeff " << i;
			throw std::exception(); // std::exception(errMsg.str().c_str());
        }
        p++;
    }
}

static void opt_inter_matrix(const char *arg, Options& options)
{
    options.inter_matrix = (uint16_t*)av_mallocz(sizeof(uint16_t) * 64);
    parse_matrix_coeffs(options.inter_matrix, arg);
}

static void opt_intra_matrix(const char *arg, Options& options)
{
    options.intra_matrix = (uint16_t*)av_mallocz(sizeof(uint16_t) * 64);
    parse_matrix_coeffs(options.intra_matrix, arg);
}


static void opt_target(const char *arg, Options& options)
{
    int norm = -1;
    static const char *const frame_rates[] = {"25", "30000/1001", "24000/1001"};

    if(!strncmp(arg, "pal-", 4)) {
        norm = 0;
        arg += 4;
    } else if(!strncmp(arg, "ntsc-", 5)) {
        norm = 1;
        arg += 5;
    } else if(!strncmp(arg, "film-", 5)) {
        norm = 2;
        arg += 5;
    } else {
        int fr;
        /* Calculate FR via float to avoid int overflow */
        fr = (int)(options.frame_rate.num * 1000.0 / options.frame_rate.den);
        if(fr == 25000) {
            norm = 0;
        } else if((fr == 29970) || (fr == 23976)) {
            norm = 1;
        } else {
            /* Try to determine PAL/NTSC by peeking in the input files */
            if(options.nb_input_files) {
                int i, j;
                for(j = 0; j < options.nb_input_files; j++) {
                    for(i = 0; i < static_cast<int>(options.input_files[j]->nb_streams); i++) {
                        AVCodecContext *c = options.input_files[j]->streams[i]->codec;
                        if(c->codec_type != CODEC_TYPE_VIDEO)
                            continue;
                        fr = c->time_base.den * 1000 / c->time_base.num;
                        if(fr == 25000) {
                            norm = 0;
                            break;
                        } else if((fr == 29970) || (fr == 23976)) {
                            norm = 1;
                            break;
                        }
                    }
                    if(norm >= 0)
                        break;
                }
            }
        }
        if(options.verbose && norm >= 0)
            fprintf(stderr, "Assuming %s for target.\n", norm ? "NTSC" : "PAL");
    }

    if(norm < 0) {
        //fprintf(stderr, "Could not determine norm (PAL/NTSC/NTSC-Film) for target.\n");
        //fprintf(stderr, "Please prefix target with \"pal-\", \"ntsc-\" or \"film-\",\n");
        //fprintf(stderr, "or set a framerate with \"-r xxx\".\n");
        //exit(1);
        throw std::exception(); // std::exception("Could not determine norm (PAL/NTSC/NTSC-Film) for target");
    }

    if(!strcmp(arg, "vcd")) {

        opt_video_codec("mpeg1video", options);
        opt_audio_codec("mp2", options);
        opt_format("vcd", options);

        opt_frame_size(norm ? "352x240" : "352x288", options);
        opt_frame_rate(frame_rates[norm], options);
        opt_default("gop", norm ? "18" : "15", options);

        opt_default("b", "1150000", options);
        opt_default("maxrate", "1150000", options);
        opt_default("minrate", "1150000", options);
        opt_default("bufsize", "327680", options); // 40*1024*8;

        opt_default("ab", "224000", options);
        options.audio_sample_rate = 44100;
        options.audio_channels = 2;

        opt_default("packetsize", "2324", options);
        opt_default("muxrate", "1411200", options); // 2352 * 75 * 8;

        /* We have to offset the PTS, so that it is consistent with the SCR.
           SCR starts at 36000, but the first two packs contain only padding
           and the first pack from the other stream, respectively, may also have
           been written before.
           So the real data starts at SCR 36000+3*1200. */
        options.mux_preload= static_cast<float>((36000+3*1200) / 90000.0); //0.44
    } else if(!strcmp(arg, "svcd")) {

        opt_video_codec("mpeg2video", options);
        opt_audio_codec("mp2", options);
        opt_format("svcd", options);

        opt_frame_size(norm ? "480x480" : "480x576", options);
        opt_frame_rate(frame_rates[norm], options);
        opt_default("gop", norm ? "18" : "15", options);

        opt_default("b", "2040000", options);
        opt_default("maxrate", "2516000", options);
        opt_default("minrate", "0", options); //1145000;
        opt_default("bufsize", "1835008", options); //224*1024*8;
        opt_default("flags", "+SCAN_OFFSET", options);


        opt_default("ab", "224000", options);
        options.audio_sample_rate = 44100;

        opt_default("packetsize", "2324", options);

    } else if(!strcmp(arg, "dvd")) {

        opt_video_codec("mpeg2video", options);
        opt_audio_codec("ac3", options);
        opt_format("dvd", options);

        opt_frame_size(norm ? "720x480" : "720x576", options);
        opt_frame_rate(frame_rates[norm], options);
        opt_default("gop", norm ? "18" : "15", options);

        opt_default("b", "6000000", options);
        opt_default("maxrate", "9000000", options);
        opt_default("minrate", "0", options); //1500000;
        opt_default("bufsize", "1835008", options); //224*1024*8;

        opt_default("packetsize", "2048", options);  // from www.mpucoder.com: DVD sectors contain 2048 bytes of data, this is also the size of one pack.
        opt_default("muxrate", "10080000", options); // from mplex project: data_rate = 1260000. mux_rate = data_rate * 8

        opt_default("ab", "448000", options);
        options.audio_sample_rate = 48000;

    } else if(!strncmp(arg, "dv", 2)) {

        opt_format("dv", options);

        opt_frame_size(norm ? "720x480" : "720x576", options);
        opt_frame_pix_fmt(!strncmp(arg, "dv50", 4) ? "yuv422p" :
                                             (norm ? "yuv411p" : "yuv420p"), options);
        opt_frame_rate(frame_rates[norm], options);

        options.audio_sample_rate = 48000;
        options.audio_channels = 2;

    } else {
        //fprintf(stderr, "Unknown target: %s\n", arg);
        //exit(1);
        throw std::exception(); // std::exception("Unknown target");
        
    }
}

static void opt_vstats_file (const char *arg, Options& options)
{
    av_free (options.vstats_filename);
    options.vstats_filename=av_strdup (arg);
}

static void opt_vstats (Options& options)
{
    char filename[40];
    time_t today2 = time(NULL);
    struct tm *today = localtime(&today2);

    snprintf(filename, sizeof(filename), "vstats_%02d%02d%02d.log", today->tm_hour, today->tm_min,
             today->tm_sec);
    opt_vstats_file(filename, options);
}

static void opt_video_bsf(const char *arg, Options& options)
{
    AVBitStreamFilterContext *bsfc= av_bitstream_filter_init(arg); //FIXME split name and args for filter at '='
    AVBitStreamFilterContext **bsfp;

    if(!bsfc){
        //fprintf(stderr, "Unknown bitstream filter %s\n", arg);
        //exit(1);
        throw std::exception(); // std::exception("Unknown bitstream filter");
    }

    bsfp= &options.video_bitstream_filters;
    while(*bsfp)
        bsfp= &(*bsfp)->next;

    *bsfp= bsfc;
}

//FIXME avoid audio - video code duplication
static void opt_audio_bsf(const char *arg, Options& options)
{
    AVBitStreamFilterContext *bsfc= av_bitstream_filter_init(arg); //FIXME split name and args for filter at '='
    AVBitStreamFilterContext **bsfp;

    if(!bsfc){
        //fprintf(stderr, "Unknown bitstream filter %s\n", arg);
        //exit(1);
        throw std::exception(); // std::exception("Unknown bitstream filter");
    }

    bsfp= &options.audio_bitstream_filters;
    while(*bsfp)
        bsfp= &(*bsfp)->next;

    *bsfp= bsfc;
}


static int opt_default(const char *opt, const char *arg, Options& options){
    int type;
    int ret = 0;
    const AVOption *o= NULL;
    int opt_types[]={AV_OPT_FLAG_VIDEO_PARAM, AV_OPT_FLAG_AUDIO_PARAM, 0, AV_OPT_FLAG_SUBTITLE_PARAM, 0};

    for(type=0; type<CODEC_TYPE_NB; type++){
        const AVOption *o2 = av_find_opt(options.avctx_opts[0], opt, NULL, opt_types[type], opt_types[type]);
        if(o2)
            ret = av_set_string3(options.avctx_opts[type], opt, arg, 1, &o);
    }
    if(!o)
        ret = av_set_string3(options.avformat_opts, opt, arg, 1, &o);
    if(!o)
        ret = av_set_string3(options.sws_opts, opt, arg, 1, &o);
    if(!o){
        if(opt[0] == 'a')
            ret = av_set_string3(options.avctx_opts[CODEC_TYPE_AUDIO], opt+1, arg, 1, &o);
        else if(opt[0] == 'v')
            ret = av_set_string3(options.avctx_opts[CODEC_TYPE_VIDEO], opt+1, arg, 1, &o);
        else if(opt[0] == 's')
            ret = av_set_string3(options.avctx_opts[CODEC_TYPE_SUBTITLE], opt+1, arg, 1, &o);
    }
    if (o && ret < 0) {
        //fprintf(stderr, "Invalid value '%s' for option '%s'\n", arg, opt);
        //exit(1);
        throw std::exception();
    }
    if(!o)
        return -1;

//    av_log(NULL, AV_LOG_ERROR, "%s:%s: %f 0x%0X\n", opt, arg, av_get_double(avctx_opts, opt, NULL), (int)av_get_int(avctx_opts, opt, NULL));

    //FIXME we should always use avctx_opts, ... for storing options so there wont be any need to keep track of whats set over this
    options.opt_names= (const char**)av_realloc(options.opt_names, sizeof(void*)*(options.opt_name_count+1));
    options.opt_names[options.opt_name_count++]= o->name;

    if(options.avctx_opts[0]->debug || options.avformat_opts->debug)
		av_log_set_level(AV_LOG_DEBUG);
    return 0;
}

const OptionDef kOptionsMap[] = {
    /* main options */
    //{ "L", 0, {(void*)show_license}, "show license" },
    //{ "h", 0, {(void*)opt_show_help}, "show help" },
    //{ "version", 0, {(void*)show_version}, "show version" },
    //{ "formats", 0, {(void*)show_formats}, "show available formats, codecs, protocols, ..." },
    { "f", HAS_ARG, {(void*)opt_format}, "force format", "fmt" },
    { "i", HAS_ARG, {(void*)opt_input_file}, "input file name", "filename" },
    { "y", 0, {(void*)opt_file_overwrite}, "overwrite output files" },
    { "map", HAS_ARG | OPT_EXPERT, {(void*)opt_map}, "set input stream mapping", "file:stream[:syncfile:syncstream]" },
    { "map_meta_data", HAS_ARG | OPT_EXPERT, {(void*)opt_map_meta_data}, "set meta data information of outfile from infile", "outfile:infile" },
    { "t", HAS_ARG, {(void*)opt_recording_time}, "set the recording time", "duration" },
    { "fs", HAS_ARG, {(void*)opt_limit_filesize}, "set the limit file size in bytes", "limit_size" }, //
    { "ss", HAS_ARG, {(void*)opt_start_time}, "set the start time offset", "time_off" },
    { "itsoffset", HAS_ARG, {(void*)opt_input_ts_offset}, "set the input ts offset", "time_off" },
    { "itsscale", HAS_ARG, {(void*)opt_input_ts_scale}, "set the input ts scale", "stream:scale" },
    { "title", HAS_ARG, {(void*)opt_str_title}, "set the title", "string" },
    { "timestamp", HAS_ARG, {(void*)opt_rec_timestamp}, "set the timestamp", "time" },
    { "author", HAS_ARG, {(void*)opt_str_author}, "set the author", "string" },
    { "copyright", HAS_ARG, {(void*)opt_str_copyright}, "set the copyright", "string" },
    { "comment", HAS_ARG, {(void*)opt_str_comment}, "set the comment", "string" },
    { "genre", HAS_ARG, {(void*)&opt_str_genre}, "set the genre", "string" },
    { "album", HAS_ARG, {(void*)opt_str_album}, "set the album", "string" },
    { "benchmark", OPT_EXPERT, {(void*)opt_do_benchmark}, "add timings for benchmarking" },
    { "dump", OPT_EXPERT, {(void*)opt_do_pkt_dump}, "dump each input packet" },
    { "hex", OPT_EXPERT, {(void*)opt_do_hex_dump}, "when dumping packets, also dump the payload" },
    { "re", OPT_EXPERT, {(void*)opt_rate_emu}, "read input at native frame rate", "" },
    { "loop_input", OPT_EXPERT, {(void*)opt_loop_input}, "loop (current only works with images)" },
    { "loop_output", HAS_ARG | OPT_EXPERT, {(void*)opt_loop_output}, "number of times to loop output in formats that support looping (0 loops forever)", "" },
    { "v", HAS_ARG, {(void*)opt_verbose}, "control amount of logging", "verbose" },
    { "target", HAS_ARG, {(void*)opt_target}, "specify target file type (\"vcd\", \"svcd\", \"dvd\", \"dv\", \"dv50\", \"pal-vcd\", \"ntsc-svcd\", ...)", "type" },
    { "threads", HAS_ARG | OPT_EXPERT, {(void*)opt_thread_count}, "thread count", "count" },
    { "vsync", HAS_ARG | OPT_EXPERT, {(void*)opt_video_sync_method}, "video sync method", "" },
    { "async", HAS_ARG | OPT_EXPERT, {(void*)opt_audio_sync_method}, "audio sync method", "" },
    { "adrift_threshold", HAS_ARG | OPT_EXPERT, {(void*)opt_audio_drift_threshold}, "audio drift threshold", "" },
    { "vglobal", HAS_ARG | OPT_EXPERT, {(void*)opt_video_global_header}, "video global header storage type", "" },
    { "copyts", OPT_EXPERT, {(void*)opt_copy_ts}, "copy timestamps" },
    { "shortest", OPT_EXPERT, {(void*)opt_shortest}, "finish encoding within shortest input" }, //
    { "dts_delta_threshold", HAS_ARG | OPT_EXPERT, {(void*)opt_dts_delta_threshold}, "timestamp discontinuity delta threshold", "" },
    { "programid", HAS_ARG | OPT_EXPERT, {(void*)&opt_programid}, "desired program number", "" },

    /* video options */
    { "vframes", HAS_ARG | OPT_VIDEO, {(void*)opt_max_frames_video}, "set the number of video frames to record", "number" },
    { "dframes", HAS_ARG, {(void*)opt_max_frames_data}, "set the number of data frames to record", "number" },
    { "r", HAS_ARG | OPT_VIDEO, {(void*)opt_frame_rate}, "set frame rate (Hz value, fraction or abbreviation)", "rate" },
    { "s", HAS_ARG | OPT_VIDEO, {(void*)opt_frame_size}, "set frame size (WxH or abbreviation)", "size" },
    { "aspect", HAS_ARG | OPT_VIDEO, {(void*)opt_frame_aspect_ratio}, "set aspect ratio (4:3, 16:9 or 1.3333, 1.7777)", "aspect" },
    { "pix_fmt", HAS_ARG | OPT_EXPERT | OPT_VIDEO, {(void*)opt_frame_pix_fmt}, "set pixel format, 'list' as argument shows all the pixel formats supported", "format" },
    { "croptop", HAS_ARG | OPT_VIDEO, {(void*)opt_frame_crop_top}, "set top crop band size (in pixels)", "size" },
    { "cropbottom", HAS_ARG | OPT_VIDEO, {(void*)opt_frame_crop_bottom}, "set bottom crop band size (in pixels)", "size" },
    { "cropleft", HAS_ARG | OPT_VIDEO, {(void*)opt_frame_crop_left}, "set left crop band size (in pixels)", "size" },
    { "cropright", HAS_ARG | OPT_VIDEO, {(void*)opt_frame_crop_right}, "set right crop band size (in pixels)", "size" },
    { "padtop", HAS_ARG | OPT_VIDEO, {(void*)opt_frame_pad_top}, "set top pad band size (in pixels)", "size" },
    { "padbottom", HAS_ARG | OPT_VIDEO, {(void*)opt_frame_pad_bottom}, "set bottom pad band size (in pixels)", "size" },
    { "padleft", HAS_ARG | OPT_VIDEO, {(void*)opt_frame_pad_left}, "set left pad band size (in pixels)", "size" },
    { "padright", HAS_ARG | OPT_VIDEO, {(void*)opt_frame_pad_right}, "set right pad band size (in pixels)", "size" },
    { "padcolor", HAS_ARG | OPT_VIDEO, {(void*)opt_pad_color}, "set color of pad bands (Hex 000000 thru FFFFFF)", "color" },
    { "intra", OPT_EXPERT | OPT_VIDEO, {(void*)opt_intra_only}, "use only intra frames"},
    { "vn", OPT_VIDEO, {(void*)opt_video_disable}, "disable video" },
    { "vdt", HAS_ARG | OPT_EXPERT | OPT_VIDEO, {(void*)opt_video_discard}, "discard threshold", "n" },
    { "qscale", HAS_ARG | OPT_EXPERT | OPT_VIDEO, {(void*)opt_qscale}, "use fixed video quantizer scale (VBR)", "q" },
    { "rc_override", HAS_ARG | OPT_EXPERT | OPT_VIDEO, {(void*)opt_video_rc_override_string}, "rate control override for specific intervals", "override" },
    { "vcodec", HAS_ARG | OPT_VIDEO, {(void*)opt_video_codec}, "force video codec ('copy' to copy stream)", "codec" },
    { "me_threshold", HAS_ARG | OPT_EXPERT | OPT_VIDEO, {(void*)opt_me_threshold}, "motion estimaton threshold",  "" },
    { "sameq", OPT_VIDEO, {(void*)opt_same_quality},
      "use same video quality as source (implies VBR)" },
    { "pass", HAS_ARG | OPT_VIDEO, {(void*)&opt_pass}, "select the pass number (1 or 2)", "n" },
    { "passlogfile", HAS_ARG | OPT_VIDEO, {(void*)opt_pass_logfilename}, "select two pass log file name", "file" },
    { "deinterlace", OPT_EXPERT | OPT_VIDEO, {(void*)opt_do_deinterlace},
      "deinterlace pictures" },
    { "psnr", OPT_EXPERT | OPT_VIDEO, {(void*)opt_do_psnr}, "calculate PSNR of compressed frames" },
    { "vstats", OPT_EXPERT | OPT_VIDEO, {(void*)&opt_vstats}, "dump video coding statistics to file" },
    { "vstats_file", HAS_ARG | OPT_EXPERT | OPT_VIDEO, {(void*)opt_vstats_file}, "dump video coding statistics to file", "file" },
#ifdef CONFIG_VHOOK
    { "vhook", HAS_ARG | OPT_EXPERT | OPT_VIDEO, {(void*)add_frame_hooker}, "insert video processing module", "module" },
#endif
    { "intra_matrix", HAS_ARG | OPT_EXPERT | OPT_VIDEO, {(void*)opt_intra_matrix}, "specify intra matrix coeffs", "matrix" },
    { "inter_matrix", HAS_ARG | OPT_EXPERT | OPT_VIDEO, {(void*)opt_inter_matrix}, "specify inter matrix coeffs", "matrix" },
    { "top", HAS_ARG | OPT_EXPERT | OPT_VIDEO, {(void*)opt_top_field_first}, "top=1/bottom=0/auto=-1 field first", "" },
    { "dc", HAS_ARG | OPT_EXPERT | OPT_VIDEO, {(void*)opt_intra_dc_precision}, "intra_dc_precision", "precision" },
    { "vtag", HAS_ARG | OPT_EXPERT | OPT_VIDEO, {(void*)opt_video_tag}, "force video tag/fourcc", "fourcc/tag" },
    { "newvideo", OPT_VIDEO, {(void*)opt_new_video_stream}, "add a new video stream to the current output stream" },
    { "qphist", OPT_EXPERT | OPT_VIDEO, { (void *)opt_qp_hist }, "show QP histogram" },

    /* audio options */
    { "aframes", HAS_ARG | OPT_AUDIO, {(void*)opt_max_frames_audio}, "set the number of audio frames to record", "number" },
    { "aq", HAS_ARG | OPT_AUDIO, {(void*)opt_audio_qscale}, "set audio quality (codec-specific)", "quality", },
    { "ar", HAS_ARG | OPT_AUDIO, {(void*)opt_audio_rate}, "set audio sampling rate (in Hz)", "rate" },
    { "ac", HAS_ARG | OPT_AUDIO, {(void*)opt_audio_channels}, "set number of audio channels", "channels" },
    { "an", OPT_AUDIO, {(void*)opt_audio_disable}, "disable audio" },
    { "acodec", HAS_ARG | OPT_AUDIO, {(void*)opt_audio_codec}, "force audio codec ('copy' to copy stream)", "codec" },
    { "atag", HAS_ARG | OPT_EXPERT | OPT_AUDIO, {(void*)opt_audio_tag}, "force audio tag/fourcc", "fourcc/tag" },
    { "vol", HAS_ARG | OPT_AUDIO, {(void*)opt_audio_volume}, "change audio volume (256=normal)" , "volume" }, //
    { "newaudio", OPT_AUDIO, {(void*)opt_new_audio_stream}, "add a new audio stream to the current output stream" },
    { "alang", HAS_ARG | OPT_AUDIO, {(void *)opt_audio_language}, "set the ISO 639 language code (3 letters) of the current audio stream" , "code" },
    { "sample_fmt", HAS_ARG | OPT_EXPERT | OPT_AUDIO, {(void*)opt_audio_sample_fmt}, "set sample format, 'list' as argument shows all the sample formats supported", "format" },

    /* subtitle options */
    { "sn", OPT_SUBTITLE, {(void*)opt_subtitle_disable}, "disable subtitle" },
    { "scodec", HAS_ARG | OPT_SUBTITLE, {(void*)opt_subtitle_codec}, "force subtitle codec ('copy' to copy stream)", "codec" },
    { "newsubtitle", OPT_SUBTITLE, {(void*)opt_new_subtitle_stream}, "add a new subtitle stream to the current output stream" },
    { "slang", HAS_ARG | OPT_SUBTITLE, {(void *)opt_subtitle_language}, "set the ISO 639 language code (3 letters) of the current subtitle stream" , "code" },

    /* grab options */
    { "vc", HAS_ARG | OPT_EXPERT | OPT_VIDEO | OPT_GRAB, {(void*)opt_video_channel}, "set video grab channel (DV1394 only)", "channel" },
    { "tvstd", HAS_ARG | OPT_EXPERT | OPT_VIDEO | OPT_GRAB, {(void*)opt_video_standard}, "set television standard (NTSC, PAL (SECAM))", "standard" },
    { "isync", OPT_EXPERT | OPT_GRAB, {(void*)opt_input_sync}, "sync read on input", "" },

    /* muxer options */
    { "muxdelay", HAS_ARG | OPT_EXPERT, {(void*)opt_mux_max_delay}, "set the maximum demux-decode delay", "seconds" },
    { "muxpreload", HAS_ARG | OPT_EXPERT, {(void*)opt_mux_preload}, "set the initial demux-decode delay", "seconds" },

    { "absf", HAS_ARG | OPT_AUDIO | OPT_EXPERT, {(void*)opt_audio_bsf}, "", "bitstream filter" },
    { "vbsf", HAS_ARG | OPT_VIDEO | OPT_EXPERT, {(void*)opt_video_bsf}, "", "bitstream filter" },

    { "default", OPT_FUNC2 | HAS_ARG | OPT_AUDIO | OPT_VIDEO | OPT_EXPERT, {(void*)opt_default}, "generic catch all option", "" },
    { NULL, },
};

void parse_arg_file(const char *filename, Options& options)
{
    opt_output_file(filename, options);
}

void transcode::SetLogFile(const std::string& filePath)
{
	LogFileOpener::SetLogFile(filePath);
}

void transcode::Transcode(const std::vector<std::string>& argv, Progress& progress)
{
    int i;
    int64_t ti;
	Options* optionsPtr = new Options();
	Options& options = *optionsPtr;

	try
	{
		av_register_all();

		for(i=0; i<CODEC_TYPE_NB; i++){
			options.avctx_opts[i]= avcodec_alloc_context2((CodecType)i);
		}
		options.avformat_opts = avformat_alloc_context();
		options.sws_opts = sws_getContext(16,16,(PixelFormat)0, 16,16,(PixelFormat)0, options.sws_flags, NULL,NULL,NULL);

		if (argv.size() == 0) {
			throw std::exception(); // std::exception("Invalid number of arguments");
		}
         
		/* parse options */
		parse_options(argv, kOptionsMap, options);

		// This will open and close the log file automatically.
		std::auto_ptr<LogFileOpener> logOpener;
		if (av_log_get_level() != AV_LOG_QUIET)
			logOpener.reset(new LogFileOpener());

		/* file converter / grab */
		if (options.nb_output_files <= 0) {
			//fprintf(stderr, "Must supply at least one output file\n");
			//exit(1);
			throw std::exception(); // std::exception("Must supply at least one output file");
		}

		if (options.nb_input_files == 0) {
			//fprintf(stderr, "Must supply at least one input file\n");
			//exit(1);
			throw std::exception(); // std::exception("Must supply at least one input file");
		}

		ti = getutime();
		if (av_encode(options.output_files, options.nb_output_files, options.input_files, options.nb_input_files,
				  options.stream_maps, options.nb_stream_maps, options, progress) < 0 )
            throw std::exception();
		ti = getutime() - ti;
		//if (do_benchmark) {
		//    printf("bench: utime=%0.3fs\n", ti / 1000000.0);
		//}


	#ifdef CONFIG_POWERPC_PERF
		extern void powerpc_display_perf_report(void);
		powerpc_display_perf_report();
	#endif /* CONFIG_POWERPC_PERF */

		if (received_sigterm) {
			//fprintf(stderr,
			//    "Received signal %d: terminating.\n",
			//    (int) received_sigterm);
			//std::stringstream errMsg;
			//errMsg << "Received signal: " << (int) received_sigterm;
			throw std::exception(); // std::exception(errMsg.str().c_str());
		}
	}
	catch (...)
	{
		if (optionsPtr != 0)
			delete optionsPtr;
		throw;
	}

	if (optionsPtr != 0)
		delete optionsPtr;
}
