/*
 * TranscodeTool.mm
 * Copyright (C) 2008 doubleTwist Corporation
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#import <Foundation/Foundation.h>
#import "TranscodeTool.h"
#import "Transcode.h"
#import "NSString+UUID.h"

class TranscodeProgress : public transcode::Progress
{
public:
    TranscodeProgress( MainThread *mainThread, NSString *outputFile );
    ~TranscodeProgress();
    
    bool Update(float percentComplete);
    
private:
    MainThread *mainThread;
    NSString *outputFile;
};

TranscodeProgress::TranscodeProgress( MainThread *mainThread, NSString *outputFile )
{
    this->mainThread = [mainThread retain];
    this->outputFile = [outputFile retain];
}

TranscodeProgress::~TranscodeProgress()
{
    [mainThread release];
    [outputFile release];
}

bool TranscodeProgress::Update(float percentComplete)
{
    return [mainThread progressUpdate:outputFile percentComplete:[NSNumber numberWithFloat:percentComplete]];
}

@protocol TranscodeToolProgressProtocol
- (bool)progressUpdateThroughConnection:(oneway NSString *)outputFile percentComplete:(oneway NSNumber *)percent;
- (void)transcodeCompletedThroughConnection:(oneway NSString *)outputFile successfully:(oneway BOOL)successfully;
@end

@interface MainThread()
- (void)transcodeFilesThread:(NSArray *)args;
@end

@implementation MainThread

- (id)initWithServerIdentifier:(NSString *)serverIdentifier parentProcessIdentifier:(int)parentProcessIdentifier
{
    self = [super init];
    if ( self != NULL )
    {
        myServerIdentifier = [serverIdentifier retain];
        myParentProcessIdentifier = parentProcessIdentifier;
    }
    return self;
}

- (void)dealloc
{
    [super dealloc];
}

- (BOOL)isParentProcessStillAlive
{
    NSArray *applications = [[NSWorkspace sharedWorkspace] launchedApplications];
    NSEnumerator *enumerator = [applications objectEnumerator];
    NSDictionary *application_info;
    while ( (application_info = [enumerator nextObject]) != NULL )
    {
        NSNumber *process_identifier_number = [application_info objectForKey:@"NSApplicationProcessIdentifier"];
        if ( process_identifier_number != NULL && [process_identifier_number intValue] == myParentProcessIdentifier )
            return YES;
    }
    return NO;
}

- (void)run
{
    NSConnection *connection = [NSConnection defaultConnection];    
    [connection setRootObject:self];
    if ( [connection registerName:myServerIdentifier] == YES )
    {
        myServerProxy = [[NSConnection rootProxyForConnectionWithRegisteredName:@"TranscodeUtility" host:nil] retain];
        if( myServerProxy != nil )
        {
            [myServerProxy setProtocolForProxy:@protocol(TranscodeProgressProtocol)];
        }
        
        NSRunLoop *run_loop = [NSRunLoop currentRunLoop];
        
        BOOL result = YES;
        BOOL is_parent_process_alive = YES;
        
        do
        {
            NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
            
            result = [run_loop runMode:NSDefaultRunLoopMode beforeDate:[NSDate dateWithTimeIntervalSinceNow:0.5]];

            is_parent_process_alive = [self isParentProcessStillAlive];
            
            [pool release];
        }
        while (is_parent_process_alive && result);
    }
    
    [myServerProxy release];
    myServerProxy = nil;
}

- (NSDictionary *)getMetadataForFile:(oneway NSString *)file
{
    NSMutableDictionary *md = [NSMutableDictionary dictionary];

    transcode::Metadata meta;
    try
    {
        meta = transcode::GetMetadata(std::string([file UTF8String]));
    }
    catch( ... )
    {
        md = nil;
    }
    
    if( md == nil )
        return nil;

    // general
    [md setValue:[NSNumber numberWithInt:meta.Type()] forKey:kMDItemMediaType];
    [md setValue:[NSNumber numberWithInt:meta.Format()] forKey:kMDItemFileFormat];
    [md setValue:[NSNumber numberWithDouble:meta.Duration()] forKey:kMDItemDuration];
    [md setValue:[NSNumber numberWithLong:meta.AverageBitrate()] forKey:kMDItemAverageBitrate];
    [md setValue:[NSNumber numberWithBool:meta.HasAudio()] forKey:kMDItemHasAudio];
    [md setValue:[NSString stringWithCString:meta.Title().c_str()] forKey:kMDItemTitle];
    [md setValue:[NSString stringWithCString:meta.Artist().c_str()] forKey:kMDItemArtist];
    [md setValue:[NSString stringWithCString:meta.Album().c_str()] forKey:kMDItemAlbum];
    [md setValue:[NSString stringWithCString:meta.Genre().c_str()] forKey:kMDItemGenre];

    // video
    NSMutableDictionary *mdv = [NSMutableDictionary dictionary];
    [mdv setValue:[NSNumber numberWithInt:meta.VideoMetaData().Encoding()] forKey:kMDEncoding];
    [mdv setValue:[NSNumber numberWithDouble:meta.VideoMetaData().Duration()] forKey:kMDDuration];
    [mdv setValue:[NSNumber numberWithLong:meta.VideoMetaData().AverageBitrate()] forKey:kMDAverageBitrate];
    [mdv setValue:[NSNumber numberWithInt:meta.VideoMetaData().Width()] forKey:kMDWidth];
    [mdv setValue:[NSNumber numberWithInt:meta.VideoMetaData().Height()] forKey:kMDHeight];
    [mdv setValue:[NSNumber numberWithDouble:meta.VideoMetaData().FrameRate()] forKey:kMDFrameRate];
    [mdv setValue:[NSNumber numberWithDouble:meta.VideoMetaData().PixelAspectRatio()] forKey:kMDPixelAspectRatio];
    [mdv setValue:[NSNumber numberWithInt:meta.VideoMetaData().Profile()] forKey:kMDProfile];
    [mdv setValue:[NSNumber numberWithInt:meta.VideoMetaData().Level()] forKey:kMDLevel];
    [md setValue:[NSDictionary dictionaryWithDictionary:mdv] forKey:kMDVideo];

    // audio
    NSMutableDictionary *mda = [NSMutableDictionary dictionary];
    [mda setValue:[NSNumber numberWithInt:meta.AudioMetaData().Encoding()] forKey:kMDEncoding];
    [mda setValue:[NSNumber numberWithDouble:meta.AudioMetaData().Duration()] forKey:kMDDuration];
    [mda setValue:[NSNumber numberWithInt:meta.AudioMetaData().Bitrate()] forKey:kMDBitrate];
    [mda setValue:[NSNumber numberWithInt:meta.AudioMetaData().SampleRate()] forKey:kMDSampleRate];
    [mda setValue:[NSNumber numberWithInt:meta.AudioMetaData().SampleSize()] forKey:kMDSampleSize];
    [mda setValue:[NSNumber numberWithInt:meta.AudioMetaData().ChannelCount()] forKey:kMDChannelCount];
    [md setValue:[NSDictionary dictionaryWithDictionary:mda] forKey:kMDAudio];
    
    return [NSDictionary dictionaryWithDictionary:md];
}

- (NSData *)getThumbnailForFile:(oneway NSString *)file usingArguments:(oneway NSArray *)arguments
{
	NSArray *files = [NSArray arrayWithObject:file];
	NSString *outputFile = [NSTemporaryDirectory() stringByAppendingPathComponent:[NSString stringWithNewGUID]];
	NSArray *args = [NSArray arrayWithObjects:[files retain], [outputFile retain], [arguments retain], nil];
	[self transcodeFilesThread:args];
	
	if( ![[NSFileManager defaultManager] fileExistsAtPath:outputFile] )
		return nil;
	
	return [[[NSData alloc] initWithContentsOfFile:outputFile] autorelease];
}

- (void)transcodeFiles:(oneway NSArray *)files toFile:(oneway NSString *)outputFile usingArguments:(oneway NSArray *)arguments
{
    NSArray *args = [NSArray arrayWithObjects:[files retain], [outputFile retain], [arguments retain], nil];
    [NSThread detachNewThreadSelector:@selector(transcodeFilesThread:) toTarget:self withObject:args];
}

- (void)transcodeFilesThread:(NSArray *)args
{    
    NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
    BOOL success = NO;
    
    NSArray *files = [args objectAtIndex:0];
    NSString *outputFile = [args objectAtIndex:1];
    NSArray *arguments = [args objectAtIndex:2];
    
    TranscodeProgress progress( self, outputFile );
    std::vector<std::string> argv;
    
    for( NSString *file in files )
    {
        argv.push_back( [@"-i" UTF8String] );
        argv.push_back( [file UTF8String] );
    }
    
    for( NSString *arg in arguments )
        argv.push_back([arg UTF8String]);
    
    argv.push_back( [outputFile UTF8String] );
    
    try
    {
        transcode::Transcode(argv, progress);
        success = YES;
    }
    catch( ... )
    {
        success = NO;
    }
    
    for( id obj in args )
        [obj release];
    
    [self transcodeCompleted:outputFile successfully:success];
    
    [pool release];
}

// this method should only be called via a connection (to ourself). used to make sure all calls
// to the transcode utility process all come from a single thread.
- (bool)progressUpdateThroughConnection:(oneway NSString *)outputFile percentComplete:(oneway NSNumber *)percent;
{
    return [(id<TranscodeProgressProtocol>)myServerProxy progressUpdate:outputFile percentComplete:percent];
}

- (bool)progressUpdate:(NSString *)outputFile percentComplete:(NSNumber *)percent
{
    if ( myServerProxy != nil )
    {        
        id proxy = [NSConnection rootProxyForConnectionWithRegisteredName:myServerIdentifier host:nil];
        
        [proxy setProtocolForProxy:@protocol(TranscodeToolProgressProtocol)];
        
        return [proxy progressUpdateThroughConnection:outputFile percentComplete:percent];
    }
    
    return NO;
}

- (void)transcodeCompletedThroughConnection:(oneway NSString *)outputFile successfully:(oneway BOOL)successfully
{
    return [(id<TranscodeProgressProtocol>)myServerProxy transcodeCompleted:outputFile successfully:successfully];
}

- (void)transcodeCompleted:(NSString *)outputFile successfully:(BOOL)successfully
{
    if ( myServerProxy != nil )
    {        
        id proxy = [NSConnection rootProxyForConnectionWithRegisteredName:myServerIdentifier host:nil];
        
        [proxy setProtocolForProxy:@protocol(TranscodeToolProgressProtocol)];
        
        return [proxy transcodeCompletedThroughConnection:outputFile successfully:successfully];
    }
}

@end

int main (int argc, const char * argv[])
{
    NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
    
    NSApplicationLoad();
    
    NSString *serverIdentifier = [NSString stringWithCString:argv[1]];
    
    int parentProcessIdentifier = atoi(argv[2]);
    
    if ( parentProcessIdentifier > 0 )
    {
        MainThread *server = [[[MainThread alloc] initWithServerIdentifier:serverIdentifier parentProcessIdentifier:parentProcessIdentifier] autorelease];
        
        [server run];
    }
    
    [pool release];
    
    return 0;
}
